# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Fast divide-and-conquer (DC) algorithms for computing Pareto frontier."""

import abc

import numpy as np


class BaseParetoOptimalAlgorithm(metaclass=abc.ABCMeta):
  """Base metaclass for calculating pareto frontiers."""

  @abc.abstractmethod
  def is_pareto_optimal(self, points: np.ndarray) -> np.ndarray:
    """An algorithm to find the Pareto frontier.

    Args:
      points: M-by-D 2D array of points. M = number of points, D = dimension.

    Returns:
      A boolean array of length m corresponding to each point in points. True
      if pareto optimal.
    """
    pass

  @abc.abstractmethod
  def is_pareto_optimal_against(self, points: np.ndarray, against: np.ndarray,
                                *, strict: bool) -> np.ndarray:
    """An algorithm to check if the points are pareto optimal or dominated.

    NOTE: p dominates q if p_i >= q_i and strictly if p_i != q_i.

    Dominated: A point is strictly dominated by some point in dominated_points.
    Equal: A non-strictly dominated point equal to some point in
    dominated_points.
    Optimal: A point is not dominated.

    Args:
      points: M-by-D 2D array of points to be checked.  M = number of points, D
        = dimension.
      against: N-by-D 2D array of dominating points. N = number of dominating
        points, D = dimension.
      strict: If true, non-strictly dominated (i.e. equal) points are considered
        optimal.

    Returns:
      A length N boolean array corresponding to each point in points. True means
      that the point is not dominated by any point in against.
    """
    pass


class NaiveParetoOptimalAlgorithm(BaseParetoOptimalAlgorithm):
  """Naive implementation of Pareto frontier calculation."""

  def is_pareto_optimal_against(self, points: np.ndarray, against: np.ndarray,
                                *, strict: bool) -> np.ndarray:
    """Naive quadratic-time implementation. See base class."""
    if against.shape[1] != points.shape[1]:
      raise ValueError(f'Shape for against {against.shape}'
                       f'does not match in last dim for points {points.shape}')
    is_optimal = np.zeros(len(points), dtype=bool)
    for i, point in enumerate(points):
      strict_dominating = np.any(point > against, axis=1)
      if np.all(strict_dominating):
        # Optimal point since no points dominate it.
        is_optimal[i] = True
      # Mark equal as optimal if strict is True.
      elif strict and np.all(strict_dominating
                             | np.all(point == against, axis=1)):
        is_optimal[i] = True

    return is_optimal

  def is_pareto_optimal(self, points: np.ndarray) -> np.ndarray:
    """Naive quadratic-time implementation. See base class."""
    is_optimal = np.ones(len(points), dtype=bool)
    for i, point in enumerate(points):
      if is_optimal[i]:
        # At each step, revise all potentially optimal points.
        is_optimal[is_optimal] = (
            np.any(points[is_optimal] > point, axis=1)
            | np.all(points[is_optimal] == point, axis=1))
    return is_optimal


class FastParetoOptimalAlgorithm(BaseParetoOptimalAlgorithm):
  """Fast DC/recursive algorithm deciding if points are optimal or dominated.

  NOTE: p dominates q if p_i >= q_i and strictly if p_i != q_i.

  Citation on first publication:
  https://static.aminer.org/pdf/PDF/000/211/201/on_the_computational_complexity_of_finding_the_maxima_of_a.pdf
  """

  def __init__(
      self,
      base_algorithm: BaseParetoOptimalAlgorithm = NaiveParetoOptimalAlgorithm(
      ),
      *,
      recursive_threshold: int = 10000):
    """Init.

    To use with XLA:
      from vizier.pyvizier.multimetric import xla_pareto

      algo =
      FastParetoOptimalAlgorithm(xla_pareto.JaxParetoOptimalAlgorithm())
      algo.is_pareto_optimal(points)

    Args:
      base_algorithm: Base pareto computation algorithm used as base case.
      recursive_threshold: If points are fewer than this threshold, use base
        case naive algorithm.
    """
    self._base_algorithm = base_algorithm
    self._recursive_threshold = recursive_threshold

  def is_pareto_optimal_against(self, points: np.ndarray, against: np.ndarray,
                                *, strict: bool) -> np.ndarray:
    """Fast DC algorithm/recursive algorithm. See base class."""
    # Edge case: When either points or against are empty.
    if len(points) <= 0:
      return np.array([], dtype=bool)
    if len(against) <= 0:
      return np.array(np.ones(len(points), dtype=bool))

    # Base case.
    if len(against) < self._recursive_threshold or len(
        points) < self._recursive_threshold:
      if strict:
        return np.array(
            self._base_algorithm.is_pareto_optimal_against(
                points, against, strict=True),
            dtype=bool)
      else:
        return np.array(
            self._base_algorithm.is_pareto_optimal_against(
                points, against, strict=False),
            dtype=bool)

    # Base case: If 1D, this is a simple linear time algorithm.
    if against.shape[1] <= 1:
      max_value = np.max(against)
      if strict:
        return (points >= max_value).squeeze()
      else:
        return (points > max_value).squeeze()

    # Sort points from lowest to highest for first dimension and find split.
    ascending_indices = (points[:, 0]).argsort()
    sorted_points = points[ascending_indices]
    split_index = round(len(points) / 2)
    split_value = sorted_points[split_index][0]
    while sorted_points[split_index][0] == split_value:
      split_index += 1
      # Hard to find a clean split. Resort to simple algorithm.
      if split_index == len(sorted_points):
        if strict:
          return np.array(
              self._base_algorithm.is_pareto_optimal_against(
                  points, against, strict=True),
              dtype=bool)
        else:
          return np.array(
              self._base_algorithm.is_pareto_optimal_against(
                  points, against, strict=False),
              dtype=bool)

    # Sort dominating points from lowest to highest for first dim and split.
    sorted_dominating = against[(against[:, 0]).argsort()]
    dominating_split_index = np.searchsorted(
        sorted_dominating[:, 0], split_value, side='right')

    # Divide points, dominated_points and then recurse.
    lower_points = sorted_points[:split_index]
    upper_points = sorted_points[split_index:]
    lower_dominating = sorted_dominating[:dominating_split_index]
    upper_dominating = sorted_dominating[dominating_split_index:]

    upper_optimal = self.is_pareto_optimal_against(
        points=upper_points, against=upper_dominating, strict=strict)
    lower_optimal = self.is_pareto_optimal_against(
        points=lower_points, against=lower_dominating, strict=strict)
    # Can remove the first dimension due to sorting criteria. Also any equal
    # point is dominated because of the sorting of the first dimension.
    cross_optimal = self.is_pareto_optimal_against(
        points=lower_points[:, 1:],
        against=upper_dominating[:, 1:],
        strict=False)

    # Combine subresults and return.
    lower_optimal = lower_optimal & cross_optimal

    is_optimal = np.array(np.zeros(len(points), dtype=bool))
    is_optimal[ascending_indices[:split_index]] = lower_optimal
    is_optimal[ascending_indices[split_index:]] = upper_optimal

    return is_optimal

  def is_pareto_optimal(self, points: np.ndarray) -> np.ndarray:
    """Fast DC algorithm to find the Pareto frontier. See base class."""

    # Base case naive algorithm for find Pareto optimality.
    if len(points) <= self._recursive_threshold:
      return np.array(
          self._base_algorithm.is_pareto_optimal(points), dtype=bool)

    # Sort from lowest to highest on first dimension and split.
    ascending_indices = (points[:, 0]).argsort()
    sorted_points = points[ascending_indices]
    split_index = round(len(points) / 2)

    # Recurse on both subarrays and check for cross domination.
    lower_array = sorted_points[:split_index]
    higher_array = sorted_points[split_index:]
    higher_pareto = self.is_pareto_optimal(higher_array)
    lower_pareto = self.is_pareto_optimal(lower_array)
    cross_check = self.is_pareto_optimal_against(
        lower_array, higher_array, strict=True)
    lower_pareto = lower_pareto & cross_check

    is_optimal = np.zeros(len(points), dtype=bool)
    is_optimal[ascending_indices[:split_index]] = lower_pareto
    is_optimal[ascending_indices[split_index:]] = higher_pareto
    return is_optimal



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Spatio-temporal converters."""

import collections
import copy
import dataclasses
from typing import Any, Dict, List, Optional, Sequence, Tuple, Union

from absl import logging
import numpy as np
from vizier import pyvizier
from vizier.pyvizier.converters import core


@dataclasses.dataclass
class TimedLabels:
  """Corresponds to a *single* Trial's (intermediate) measurements.

  Let M be the number of measurement a trial had, Then,
    time: (M, 1) array of timestamps (See StudyMerger.to_time_and_labels_array
     pydoc for details.)
    labels: (M, 1) arrays keyed by strings, corresponding to metrics.
  """
  times: np.ndarray
  labels: Dict[str, np.ndarray]


class TimedLabelsExtractor:
  """Extracts TimeLabels."""

  CUMMAX = 'cummax'
  CUMMAX_LASTONLY = 'cummax_lastonly'
  CUMMAX_FIRSTONLY = 'cummax_firstonly'
  RAW = 'raw'

  def __init__(self,
               metric_converters: Sequence[core.ModelOutputConverter],
               timestamp: str = 'steps',
               *,
               temporal_index_points: Sequence[float] = tuple(),
               value_extraction: str = 'cummax_lastonly'):
    """Init.

    Args:
      metric_converters:
      timestamp: One of ['steps', 'elapsed_secs', 'index'].
      temporal_index_points: If provided, extract at the provided temporal index
        points.
      value_extraction: One of ['raw', 'cummax','cummax_lastonly',
        'cummax_firstonly'].
        'cummax': take the best value up to the time. For example, a
          maximization metric raw values
          (2, 1, 0, 3, 3, 2, 4, 2, 1) will be converted to: (2, 2, 2, 3, 3, 3,
            4, 4, 4).
        'cummax_lastonly': Every time there is an improvement, record the
          measurement before that improvement. For example, a maximization
          metric raw values
          (2, 1, 0, 3, 3, 2, 4, 2, 1) will be converted to: (_, _, 2, _, _, 3,
            _, _, 4).
        'cummax_firstonly':  Every time there is an improvement, record that
          value. Discard the rest except the very last measurement, so that
          stopping algorithms don't keep running a trial that's stuck in a
          plateau. For example, a maximization metric raw values
          (2, 1, 0, 3, 3, 2, 4, 2, 1) will be converted to: (2, _, _, 3, _, _,
            4, _, 4).
    """
    self.metric_converters = metric_converters
    self.temporal_index_points = np.asarray(temporal_index_points).squeeze()
    self.value_extraction = value_extraction
    self.timestamp = timestamp

    if value_extraction not in (self.CUMMAX, self.RAW, self.CUMMAX_LASTONLY,
                                self.CUMMAX_FIRSTONLY):
      raise ValueError(
          'Bad value for value_extraction rule: {}'.format(value_extraction))
    if value_extraction in [self.CUMMAX_LASTONLY, self.CUMMAX_FIRSTONLY]:
      if len(metric_converters) > 1:
        raise ValueError(
            '{} mode supports only a single metric.'.format(value_extraction))
      if self.temporal_index_points.size > 1:
        raise ValueError(
            '{} mode does not support fixed temporal_index_points.'.format(
                value_extraction))

  def _cummax_fn(self, metric_converter: core.ModelOutputConverter) -> Any:
    if (metric_converter.metric_information.goal ==
        pyvizier.ObjectiveMetricGoal.MAXIMIZE):
      return np.maximum
    else:
      return np.minimum

  def convert(self, trials: Sequence[pyvizier.Trial]) -> List[TimedLabels]:
    """Converts each trial into TimedLabels object."""
    timedlabels = []
    if self.temporal_index_points.size == 0:
      # Default mode. Temporal index points are not specified. Take all
      # available measurements.
      for t in trials:
        times = self.to_timestamps(t.measurements)
        labels = dict()
        for mc in self.metric_converters:
          # Process one metric at a time, to keep things simple.
          if self.value_extraction == self.CUMMAX:
            labels[mc.metric_information.name] = self._cummax_fn(mc).accumulate(
                mc.convert(t.measurements), axis=0)
          elif self.value_extraction in [
              self.CUMMAX_LASTONLY, self.CUMMAX_FIRSTONLY
          ]:
            # Squeeze into 1D to make indexing easier.
            marr = self._cummax_fn(mc).accumulate(
                mc.convert(t.measurements), axis=0).reshape(-1)
            if marr.size > 0:
              # Take the first and last index and the indices
              # where there was a delta.
              if self.value_extraction == self.CUMMAX_LASTONLY:
                delta_idx = np.concatenate(
                    [marr[:-1] < marr[1:],
                     np.array([True])])
              else:
                delta_idx = np.concatenate(
                    [np.array([True]), marr[:-1] < marr[1:]])
                delta_idx[-1] = True
            else:
              # No measurements were found.
              delta_idx = np.array([], dtype=bool)
            # Filter labels and times.
            labels[mc.metric_information.name] = marr[delta_idx][:, np.newaxis]
            times = times[delta_idx]
          else:
            assert self.value_extraction == self.RAW, 'Unreachable.'
            labels[mc.metric_information.name] = mc.convert(t.measurements)
        timedlabels.append(TimedLabels(times, labels))
    elif self.temporal_index_points.size and self.value_extraction == self.RAW:
      num_empty_trials = 0
      for t in trials:
        # Apply masks to only take observations at the temporal points.
        times = self.to_timestamps(t.measurements)  # shape T x 1
        mask = np.isin(times, self.temporal_index_points).squeeze()  # shape T
        if np.sum(mask) == 0:
          num_empty_trials += 1

        # Apply the mask and extract the labels.
        measurements = np.array(t.measurements)[mask]
        times = times[mask]
        labels = dict()
        for mc in self.metric_converters:
          labels[mc.metric_information.name] = mc.convert(measurements)
        timedlabels.append(TimedLabels(times, labels))
      if num_empty_trials:
        logging.warning(
            '%s Out of %s trials had zero measurement after masking.',
            num_empty_trials, len(trials))
    else:
      assert self.value_extraction == self.CUMMAX, 'Unreachable.'
      for trial in trials:
        times = self.to_timestamps(trial.measurements)  # shape T x 1

        # mask[i] is the index of the first measurement that is taken after the
        # temporal_index_points[i].
        mask = []
        for t in self.temporal_index_points:
          greater_indices = np.where(t >= times)
          if not greater_indices:
            mask.append(0)
          else:
            mask.append(greater_indices[-1])
        mask = np.asarray(mask)

        # Extract the labels and apply the mask.
        labels = dict()
        for mc in self.metric_converters:
          labels[mc.metric_information.name] = self._cummax_fn(mc).accumulate(
              mc.convert(trial.measurements), axis=0)[mask]
        timedlabels.append(TimedLabels(times, labels))

    return timedlabels

  def to_timestamps(self,
                    measurements: Sequence[pyvizier.Measurement]) -> np.ndarray:
    """"Returns an arry of shape (len(measurements), 1)."""
    timestamps = []

    for idx, measurement in enumerate(measurements):
      if self.timestamp == 'steps':
        timestamps.append(measurement.steps)
      if self.timestamp == 'elapsed_secs':
        timestamps.append(measurement.elapsed_secs)
      if self.timestamp == 'index':
        timestamps.append(idx)
    return np.asarray(timestamps)[:, np.newaxis]

  def extract_all_timestamps(
      self, trials: Sequence[pyvizier.Trial]) -> Sequence[float]:
    """Returns a sorted list of unique temporal indices occurring in trials.

    This is a cheaper alternative to:
    ```
    ts = np.concatenate([np.asarray(tl.times).flatten() for tl in timed_labels])
    return sorted(list(set(ts)))
    ```
    because it skips the label conversions.

    Args:
      trials:

    Returns:
      Sorted list of temporal indices.
    """
    all_timestamps = np.concatenate(
        [self.to_timestamps(t.measurements).flatten() for t in trials])
    return sorted(list(set(all_timestamps)))


class SparseSpatioTemporalConverter(core.TrialToNumpyDict):
  """Optimized for when the temporal dimensions are not well-aligned.

  This converter inherits from the `TrialToNumpyDict` abstractions. Timestamp
  becomes an extra feature dimension.

  For inference, however, the user has to provide `temporal_index_points` to
  specify where to predict a Trial at.
  """

  def __init__(self,
               parameter_converters: Sequence[core.DefaultModelInputConverter],
               timed_labels_extractor: TimedLabelsExtractor):
    self.trial_converter = core.DefaultTrialConverter(
        parameter_converters, timed_labels_extractor.metric_converters)
    self.timed_labels_extractor = timed_labels_extractor

  def to_xy(
      self, trials: Sequence[pyvizier.Trial]
  ) -> Tuple[Dict[str, np.ndarray], Dict[str, np.ndarray]]:
    """Returned values can be used as `x` and `y` for keras.Model.fit().

    Args:
      trials:

    Returns:
      Tuple. Both have length equal to the total number of measurements.
      x values have an extra element of (None, 1). y values are the same as
      metric_converters.
    """
    timed_labels = self.timed_labels_extractor.convert(trials)
    all_features = dict()
    all_labels = dict()
    for trial, tl in zip(trials, timed_labels):
      # Let T_i be the number of temporal observations for this trial.
      # tl.times and tl.labels have shape [T_i, 1]
      features = self.to_features(trial, tl.times)
      # Add to the return values.
      for k in features:
        if k in all_features:
          all_features[k] = np.concatenate([all_features[k], features[k]],
                                           axis=0)
        else:
          all_features[k] = features[k]

      labels = tl.labels
      for k in labels:
        if k in all_labels:
          all_labels[k] = np.concatenate([all_labels[k], labels[k]], axis=0)
        else:
          all_labels[k] = labels[k]

    return all_features, all_labels

  def to_features(self, trial, temporal_index_points) -> Dict[str, np.ndarray]:
    """Converts a trial at the specified index points.

    Args:
      trial:
      temporal_index_points:

    Returns:
      Each array has shape (B,_) where _ is determined by injected
      ModelInputConverters and B = len(trials) * len(temporal_index_points).
      In addition to what ModelInputConverters return, there is an extra array
      of shape (B, 1) corresponding to the time stamps.
    """
    features = self.trial_converter.to_features(
        [trial])  # Dict values have shape [1,1]
    # Repeat the features T_i times. Dict values have shape [T_i,1]
    features = {
        k: np.tile(v, (len(temporal_index_points), 1))
        for k, v in features.items()
    }

    # Add the timestamp feature.
    features[self.timed_labels_extractor.timestamp] = temporal_index_points
    return features

  @property
  def features_shape(self) -> Dict[str, Sequence[Union[int, None]]]:
    """Returned value can be used as `input_shape` for keras.Model.build()."""
    shapes = copy.deepcopy(self.trial_converter.features_shape)
    shapes[self.timed_labels_extractor.timestamp] = (None, 1)
    return shapes

  @property
  def output_specs(self) -> Dict[str, core.NumpyArraySpec]:
    specs = copy.deepcopy(self.trial_converter.output_specs)
    name = self.timed_labels_extractor.timestamp
    # Can't use float32 max, because
    # isinstance(np.finfo(np.float32).max, float) == False
    specs[name] = core.NumpyArraySpec.from_parameter_config(
        pyvizier.ParameterConfig.factory(
            name=name, bounds=(0.0, np.finfo(float).max)),
        core.NumpyArraySpecType.default_factory)
    return specs

  @property
  def labels_shape(self) -> Dict[str, Sequence[Union[int, None]]]:
    return self.trial_converter.labels_shape

  @property
  def metric_information(self) -> Dict[str, pyvizier.MetricInformation]:
    return self.trial_converter.metric_information


class DenseSpatioTemporalConverter(core.TrialToNumpyDict):
  """Optimized for when the temporal dimensions are well-aligned.

  This converter inherits from the `TrialToNumpyDict` abstraction. Each label
  in the dict has shape [number of trials, number of temproal index points].
  To operate in this mode, temporal_index_points must be specified.

  We can also use `to_xty()` instead of `to_xy()` in an alternative workflow
  that does not depend on the `TrialToNumpyDict` interface.
  """

  def __init__(self,
               parameter_converters: Sequence[core.DefaultModelInputConverter],
               timed_labels_extractor: TimedLabelsExtractor,
               temporal_index_points: Optional[np.ndarray] = None):
    self.trial_converter = core.DefaultTrialConverter(
        parameter_converters, timed_labels_extractor.metric_converters)
    self.timed_labels_extractor = timed_labels_extractor
    if temporal_index_points is None:
      logging.info('Empty temporal_index_points.')
      self.temporal_index_points = np.zeros([0])
    else:
      self.temporal_index_points = temporal_index_points

  def _single_timedlabels_to_temporal_observations(
      self, timed_labels: TimedLabels,
      ts: Union[Sequence[float], Sequence[int]]) -> Dict[str, np.ndarray]:
    """Subroutine of _to_temporal_observations().

    Args:
      timed_labels: A single TimedLabels object.
      ts: Length T sequence of timestamps.

    Returns:
      Dict of metric names to array of shape (T).
    """

    # Reshape things into 1-D array, so we can iterate through them like
    # python list.
    ts = np.asarray(ts).reshape([-1])
    if ts.size == 0:
      # Caller asked for nothing.
      return {k: np.array([]) for k in timed_labels.labels}

    observed_times, labels_dict = timed_labels.times.reshape(
        [-1]), timed_labels.labels
    if observed_times.size == 0:
      # Nothing was observed.
      return {k: np.zeros_like(ts) * np.nan for k in timed_labels.labels}

    query_time_iter = iter(ts)
    query_time = next(query_time_iter)
    observed_time_iter = iter(enumerate(observed_times))
    idx, observed_time = next(observed_time_iter)
    this_labels = collections.defaultdict(list)

    while query_time is not None:
      if query_time == observed_time:
        # Times match. add it.
        for key in labels_dict:
          this_labels[key].append(labels_dict[key][idx, 0])
        query_time = next(query_time_iter, None)
        idx, observed_time = next(observed_time_iter, (None, None))
      elif observed_time is None or observed_time > query_time:
        # query_time is behind, or we've scanned through all observations.
        # In either case, we don't have observed data at query_time.
        # Fill with nan and advance query_time.
        query_time = next(query_time_iter, None)
        for key in labels_dict:
          this_labels[key].append(np.nan)
      else:
        # query_time is ahead.
        idx, observed_time = next(observed_time_iter, (None, None))
    return {k: np.asarray(v) for k, v in this_labels.items()}

  def _to_temporal_observations(
      self, timed_labels_sequence: Sequence[TimedLabels],
      ts: Union[Sequence[float], Sequence[int]]) -> Dict[str, np.ndarray]:
    """Returns a dict of np arrays of temporal observations.

    Filters timed_labels.labels to leave only the time indices that appear in
    ts.

    Args:
      timed_labels_sequence: Length B sequence.
      ts: Length T sequence.

    Returns:
      Dict of metric names to array of shape (B, T).
    """

    all_labels = collections.defaultdict(lambda: np.zeros(shape=(0, len(ts))))

    for tl in timed_labels_sequence:
      labels = self._single_timedlabels_to_temporal_observations(tl, ts)
      # labels have shape [T]. Add a new axis to make it (1,T) so we
      # can concat.
      labels = {k: v[np.newaxis, :] for k, v in labels.items()}
      for k in labels:
        all_labels[k] = np.concatenate([all_labels[k], labels[k]], axis=0)
    return all_labels

  def to_xy(
      self, trials: Sequence[pyvizier.Trial]
  ) -> Tuple[Dict[str, np.ndarray], Dict[str, np.ndarray]]:
    """Returned value can be used as `x`, `y` for keras.Model.fit()."""
    all_timed_labels = self.timed_labels_extractor.convert(trials)
    labels = self._to_temporal_observations(all_timed_labels,
                                            self.temporal_index_points)

    return self.trial_converter.to_features(trials), labels

  @property
  def features_shape(self) -> Dict[str, Sequence[Union[int, None]]]:
    """Returned value can be used as `input_shape` for keras.Model.build()."""
    return self.trial_converter.features_shape

  @property
  def output_specs(self) -> Dict[str, core.NumpyArraySpec]:
    return self.trial_converter.output_specs

  @property
  def labels_shape(self) -> Dict[str, Sequence[Union[int, None]]]:
    shapes = dict()
    for mc in self.trial_converter.metric_converters:
      shapes[mc.metric_information.name] = (None,
                                            len(self.temporal_index_points))
    return shapes

  @property
  def metric_information(self) -> Dict[str, pyvizier.MetricInformation]:
    return self.trial_converter.metric_information

  def to_features(self,
                  trials: Sequence[pyvizier.Trial]) -> Dict[str, np.ndarray]:
    return self.trial_converter.to_features(trials)

  def to_xty(
      self, trials: Sequence[pyvizier.Trial], temporal_selection: str = 'auto'
  ) -> Tuple[Dict[str, np.ndarray], np.ndarray, Dict[str, np.ndarray]]:
    """Returns x, t, and y.

    Args:
      trials:
      temporal_selection: 'infer', 'default', 'auto'. If 'infer', values are
        auto-inferred based on timestamps that exist in `trials`. If 'default',
        use the class-wide `self.temporal_index_points`. If 'auto', auto-infer
        only if necessary.

    Returns:
      3-tuple of (input, temporal_index_points, observations).
      input: A Dict whose values correspond to parameters.
      temporal_index_points: 1-D array of temporal index points. If provided
        as input, then it's the same as the input value.
      observations: Dict of length equal to metrics, whose values
        have shape [len(trials), len(temporal_index_points)]. May contain NaNs.
    """
    timed_labels: List[TimedLabels] = (
        self.timed_labels_extractor.convert(trials))

    if temporal_selection == 'default' or (temporal_selection == 'auto' and
                                           self.temporal_index_points.size):
      temporal_index_points = self.temporal_index_points
    elif temporal_selection == 'infer' or (temporal_selection == 'auto' and
                                           not self.temporal_index_points.size):
      temporal_index_points = sorted(
          list(
              set(
                  np.concatenate([
                      np.asarray(tl.times).flatten() for tl in timed_labels
                  ]))))
    else:
      raise ValueError('Invalid value for temporal_index_points: '
                       f'{temporal_selection}.')

    observations = self._to_temporal_observations(timed_labels,
                                                  temporal_index_points)
    inputs = self.to_features(trials)
    return inputs, np.asarray(temporal_index_points), observations



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Hypervolume calculation (stochastic approximation) functions."""
import math
from typing import Optional, Callable


import numpy as np


def _cum_hypervolume_origin(points: np.ndarray,
                            vectors: np.ndarray) -> np.ndarray:
  """Returns a randomized approximation of the cumulative dominated hypervolume.

  See Section 3, Lemma 5 of https://arxiv.org/pdf/2006.04655.pdf for a fuller
  explanation of the technique. This assumes the reference point is the
  origin.

  NOTE: This returns an unnormalized hypervolume.

  Args:
    points: Any set of points with shape (num_points, dimension).
    vectors: Set of vectors with shape (num_vectors, dimension).

  Returns:
    Approximated cumulative dominated hypervolume of points[:i].

  Raises:
    ValueError: Points and vectors do not match in dimension.
  """
  if vectors.shape[1] != points.shape[1]:
    raise ValueError(f'Vectors shape {vectors.shape} do not match dimension of'
                     f' (second value in tuple of points shape {points.shape}')
  num_points, dimension = points.shape
  num_vectors = vectors.shape[0]
  temp_points = np.broadcast_to(points[np.newaxis, :, :],
                                [num_vectors, num_points, dimension])
  vectors = vectors.reshape([num_vectors, 1, dimension])
  # Here, ratios[i][j][k] is the kth coordinate of the jth point / ith vector.
  # Since points is (num_vectors, num_points, dimension) and vectors is
  # (num_vectors, 1, dimension), note that ratios is
  # (num_vectors, num_points, dimension).
  ratios = temp_points / vectors

  # These calculations are from Lemma 5 in above cited paper (dimension axis).
  coordinate_min_ratio = np.min(ratios, axis=2)

  # Maximizing across all points (num_points axis).
  point_max_ratio = np.maximum.accumulate(coordinate_min_ratio, axis=1)
  # Averaging across the vector axis.
  return np.mean(point_max_ratio**dimension, axis=0)


class ParetoFrontier:
  """Calculate hypervolume approximations for a Pareto frontier/set."""

  def __init__(
      self,
      points: np.ndarray,
      origin: np.ndarray,
      num_vectors: int = 10000,
      cum_hypervolume_base: Callable[[np.ndarray, np.ndarray],
                                     np.ndarray] = _cum_hypervolume_origin):
    """Takes a set of points and initializes approximating vectors.

    To use with XLA:
      from vizier.pyvizier.multimetric import xla_pareto

      front = ParetoFrontier(points, origin,
        cum_hypervolume_base = xla_pareto.jax_cum_hypervolume_origin)
      front.hypervolume(is_cumulative=True)

    Args:
      points: Any set of points with shape (num_points, dimension).
      origin: An point (1D or 2D array) with length = dimension from which
        hypervolume is computed.
      num_vectors: Number of random vectors used to approximate hypervolume.
      cum_hypervolume_base: The base algorithm used to calculate hypervolume
      from the origin. Parameters are [points, vectors].

    Raises:
      ValueError: When dimensions are mismatched between points and origin.
    """
    self._points = points
    if points.shape[1] != len(origin):
      raise ValueError(
          f'Dimension mismatch frontier points {self._points.shape}'
          f' and origin with length {len(origin)}')
    self._origin = origin
    self._cum_hypervolume_base = cum_hypervolume_base
    # Generating random vectors in the positive orthant for approximation.
    self._vectors = abs(np.random.normal(size=(num_vectors, len(origin))))
    self._vectors /= np.linalg.norm(self._vectors, axis=1)[..., np.newaxis]

  def hypervolume(self,
                  additional_points: Optional[np.ndarray] = None,
                  is_cumulative: bool = False,
                  num_shards: int = 10) -> np.ndarray:
    """Returns a randomized approximation of the dominated hypervolume.

    See Section 3, Lemma 5 of https://arxiv.org/pdf/2006.04655.pdf for a fuller
    explanation of the technique.

    Args:
      additional_points: Additional 2D array to add to hypervolume computation.
      is_cumulative: If true, a cumulative hypervolume is returned. A cumulative
      hypervolume is the a vector whose i-th entry is the running hypervolume of
      the first i-th points of points + additional_points.
      num_shards: Number of shards for breaking up self._vectors for reduced
      memory. A larger number means lower memory.

    Returns:
      Approximated dominated hypervolume of frontier and additional points.

    Raises:
      ValueError: Frontier and additional points do not match in dimension.
    """
    if additional_points is None:
      points = self._points
    else:
      if self._points.shape[1] != additional_points.shape[1]:
        raise ValueError(
            f'Dimension mismatch frontier points {self._points.shape}'
            f' and additional points {additional_points.shape}')
      points = np.concatenate((self._points, additional_points), axis=0)

    # Shift points by the origin and remove all dominated/negative points.
    points = points - self._origin
    non_positive_points = np.invert(np.all(points > 0, axis=1))
    points[non_positive_points] = 0
    if not points.size:
      return np.asarray([0.0])

    # We shard the vectors for better reduced memory usage.
    idx = np.linspace(0, len(self._vectors), num_shards + 1).astype(np.int32)
    idx = list(reversed(idx))
    # We average each vector's hypervolume estimate for a final estimate. More
    # vectors mean a more accurate estimate. (vector axis average)
    approx_hypervolume = 0
    _, dimension = points.shape
    unit_hypersphere_volume = math.pi**(
        dimension / 2) / math.gamma(dimension / 2 + 1) / 2**dimension
    for begin, end in zip(idx[1:], idx[:-1]):
      vectors = self._vectors[begin:end, :]
      # Apply maximization because Jax -> NP conversion can be imprecise.
      approx_hypervolume += np.maximum.accumulate(
          self._cum_hypervolume_base(points, vectors) * unit_hypersphere_volume
      )

    if is_cumulative:
      return np.array(approx_hypervolume / num_shards)
    else:
      return np.array(np.max(approx_hypervolume / num_shards))



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Useful assertions for testing."""
from typing import Any, Mapping

import numpy as np


def assert_arraytree_allclose(
    d1: Mapping[str, Any], d2: Mapping[str, Any], **kwargs
) -> None:
  """Assertion function for comparing two (nested) dictionaries."""
  np.testing.assert_equal(d1.keys(), d2.keys())

  for k, v in d1.items():
    if isinstance(v, dict):
      assert_arraytree_allclose(v, d2[k], **kwargs)
    else:
      try:
        np.testing.assert_allclose(v, d2[k], **kwargs)
      except TypeError as e:
        np.testing.assert_equal(
            v, d2[k], err_msg=f'Using assert_equal due to typeerror: {e}.'
        )



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import copy

from vizier import pyvizier as vz
from vizier._src.benchmarks.experimenters import noisy_experimenter
from vizier._src.benchmarks.experimenters import numpy_experimenter
from vizier._src.benchmarks.experimenters import sign_flip_experimenter
from vizier._src.benchmarks.experimenters.synthetic import bbob

from absl.testing import absltest
from absl.testing import parameterized


# TODO: Add multimetric test once we have a multi-objective
# experimenter.
class SignFlipExperimenterTest(parameterized.TestCase):

  @parameterized.parameters((True,), (False))
  def test_flipped_bbob(self, flip_objectives_only: bool):
    exptr = numpy_experimenter.NumpyExperimenter(
        bbob.Sphere, bbob.DefaultBBOBProblemStatement(2)
    )
    exptr = noisy_experimenter.NoisyExperimenter(
        exptr, noise_fn=lambda v: v - 1
    )
    flipped_exptr = sign_flip_experimenter.SignFlipExperimenter(
        exptr, flip_objectives_only
    )

    metric_name = exptr.problem_statement().single_objective_metric_name
    self.assertEqual(
        flipped_exptr.problem_statement().metric_information.item().goal,
        vz.ObjectiveMetricGoal.MAXIMIZE,
    )

    suggestion_for_original = vz.Trial(parameters={'x0': 0.2, 'x1': -3.2})
    suggestion_for_flipped = copy.deepcopy(suggestion_for_original)

    exptr.evaluate([suggestion_for_original])
    flipped_exptr.evaluate([suggestion_for_flipped])

    self.assertEqual(
        suggestion_for_original.final_measurement.metrics[metric_name].value,
        -1.0
        * suggestion_for_flipped.final_measurement.metrics[metric_name].value,
    )

    aux_metric_name = metric_name + '_before_noise'
    if flip_objectives_only:
      self.assertEqual(
          suggestion_for_original.final_measurement.metrics[
              aux_metric_name
          ].value,
          suggestion_for_flipped.final_measurement.metrics[
              aux_metric_name
          ].value,
      )
    else:
      self.assertEqual(
          suggestion_for_original.final_measurement.metrics[
              aux_metric_name
          ].value,
          -1.0
          * suggestion_for_flipped.final_measurement.metrics[
              aux_metric_name
          ].value,
      )


if __name__ == '__main__':
  absltest.main()



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Tests for designer_policy."""

from typing import Optional, Sequence
from vizier import algorithms as vza
from vizier import pythia
from vizier import pyvizier as vz
from vizier._src.algorithms.policies import designer_policy as dp
from vizier.interfaces import serializable
from absl.testing import absltest


class _FakeDesigner(vza.Designer):

  def __init__(self):
    self.num_incorporated_completed_trials = 0
    self._num_incorporated_active_trials = 0
    self.last_completed = []
    self.last_active = []

  def suggest(
      self, count: Optional[int] = None
  ) -> Sequence[vz.TrialSuggestion]:
    return [vz.TrialSuggestion(vz.ParameterDict())] * count

  def update(
      self, completed: vza.CompletedTrials, all_active: vza.ActiveTrials
  ) -> None:
    self.last_completed = completed.trials
    self.last_active = all_active.trials
    self.num_incorporated_completed_trials += len(completed.trials)
    self._num_incorporated_active_trials += len(all_active.trials)


class _FakeSerializableDesigner(vza.PartiallySerializableDesigner,
                                vza.SerializableDesigner):

  def __init__(self):
    self.num_incorporated_completed_trials = 0
    self.num_incorporated_active_trials = 0
    self.last_completed = []
    self.last_active = []

    # For debugging and testing. Not stored in the metadata.
    self.last_completed = None

  def suggest(self,
              count: Optional[int] = None) -> Sequence[vz.TrialSuggestion]:
    return [vz.TrialSuggestion(vz.ParameterDict())] * count

  def update(
      self, completed: vza.CompletedTrials, all_active: vza.ActiveTrials
  ):
    self.last_completed = completed.trials
    self.last_active = all_active.trials
    self.num_incorporated_completed_trials += len(completed.trials)
    self.num_incorporated_active_trials += len(all_active.trials)

  def dump(self) -> vz.Metadata:
    md = vz.Metadata()
    md.ns('ns1')['foo1'] = 'bar1'
    md.ns('ns1').ns('ns11')['foo11'] = 'bar11'
    md['num_incorporated_completed_trials'] = str(
        self.num_incorporated_completed_trials
    )
    return md

  @classmethod
  def recover(cls, md: vz.Metadata) -> '_FakeSerializableDesigner':
    try:
      designer = cls()
      designer.num_incorporated_completed_trials = int(
          md['num_incorporated_completed_trials']
      )
      return designer
    except KeyError as e:
      raise KeyError(f'Cannot find in {md}') from e

  def load(self, md: vz.Metadata):
    try:
      self.num_incorporated_completed_trials = int(
          md['num_incorporated_completed_trials']
      )
    except KeyError as e:
      raise serializable.DecodeError(f'Cannot find in {md}') from e


_NUM_INITIAL_COMPLETED_TRIALS = 10
_NUM_INITIAL_ACTIVE_TRIALS = 2


def _create_runner() -> pythia.InRamPolicySupporter:
  """Creates the default runner with completed and active trials."""
  runner = pythia.InRamPolicySupporter(vz.ProblemStatement())
  runner.AddTrials(
      [
          vz.Trial().complete(vz.Measurement())
          for _ in range(_NUM_INITIAL_COMPLETED_TRIALS)
      ]
  )
  # The default status is ACTIVE.
  runner.AddTrials([vz.Trial() for _ in range(_NUM_INITIAL_ACTIVE_TRIALS)])
  return runner


class DesignerPolicyNormalOperationTest(absltest.TestCase):
  """Tests Designer policies under error-free conditions."""

  def setUp(self):
    super().setUp()
    self.maxDiff = None

  def test_restore_fully_serializable_designer(self):
    runner = _create_runner()
    policy = dp.SerializableDesignerPolicy(
        problem_statement=vz.ProblemStatement(),
        supporter=runner,
        designer_factory=lambda _, **kwargs: _FakeSerializableDesigner(),
        designer_cls=_FakeSerializableDesigner,
        ns_root='test',
        verbose=2,
    )
    runner.SuggestTrials(policy, 5)
    # Simulate restoring the designer from metadata.
    metadata = policy.dump()
    restored_policy = dp.SerializableDesignerPolicy(
        problem_statement=vz.ProblemStatement(metadata=metadata),
        supporter=runner,
        designer_factory=lambda _, **kwargs: _FakeSerializableDesigner(),
        designer_cls=_FakeSerializableDesigner,
        ns_root='test',
        verbose=2,
    )
    restored_policy.load(metadata)
    designer = restored_policy.designer
    self.assertEqual(
        getattr(designer, 'num_incorporated_completed_trials'),
        _NUM_INITIAL_COMPLETED_TRIALS,
    )

  def test_restore_partially_serializable_designer(self):
    runner = _create_runner()
    policy = dp.PartiallySerializableDesignerPolicy(
        vz.ProblemStatement(),
        runner,
        lambda _, **kwargs: _FakeSerializableDesigner(),
        ns_root='test',
        verbose=2,
    )
    runner.SuggestTrials(policy, 5)
    metadata = policy.dump()
    # Simluate restoring the policy from the metadata.
    restored_policy = dp.PartiallySerializableDesignerPolicy(
        vz.ProblemStatement(),
        runner,
        lambda _, **kwargs: _FakeSerializableDesigner(),
        ns_root='test',
        verbose=2,
    )
    restored_policy.load(metadata)
    self.assertLen(
        restored_policy._cache._incorporated_completed_trial_ids,
        _NUM_INITIAL_COMPLETED_TRIALS,
    )
    self.assertEqual(
        getattr(restored_policy.designer, 'num_incorporated_completed_trials'),
        _NUM_INITIAL_COMPLETED_TRIALS,
    )

  def test_update_stateless_designer(self):
    runner = _create_runner()
    designer = _FakeDesigner()
    policy = dp.DesignerPolicy(runner, lambda _, **kwargs: designer)
    runner.SuggestTrials(policy, 5)
    self.assertLen(
        designer.last_completed,
        _NUM_INITIAL_COMPLETED_TRIALS,
    )
    self.assertLen(
        designer.last_active,
        _NUM_INITIAL_ACTIVE_TRIALS,
    )
    # Add newly completed trials
    runner.AddTrials([vz.Trial().complete(vz.Measurement()) for _ in range(33)])
    runner.SuggestTrials(policy, 15)
    # Check that all completed trials are passed again.
    self.assertLen(
        designer.last_completed,
        _NUM_INITIAL_COMPLETED_TRIALS + 33,
    )
    # Check that all active trials are passed again, including the new ones.
    self.assertLen(
        designer.last_active,
        _NUM_INITIAL_ACTIVE_TRIALS + 5,
    )

  def test_update_stateful_designer(self):
    runner = _create_runner()
    designer = _FakeSerializableDesigner()
    policy = dp.PartiallySerializableDesignerPolicy(
        vz.ProblemStatement(),
        runner,
        lambda _, **kwargs: designer,
        ns_root='test',
        verbose=2,
    )
    runner.SuggestTrials(policy, 11)
    self.assertLen(
        designer.last_completed,
        _NUM_INITIAL_COMPLETED_TRIALS,
    )
    self.assertLen(
        designer.last_active,
        _NUM_INITIAL_ACTIVE_TRIALS,
    )
    # Add newly completed trials
    runner.AddTrials([vz.Trial().complete(vz.Measurement()) for _ in range(33)])
    runner.SuggestTrials(policy, 5)
    # Check that completed trials are not passed again.
    self.assertLen(designer.last_completed, 33)
    # Check that the newly active trials are passed.
    self.assertLen(designer.last_active, _NUM_INITIAL_ACTIVE_TRIALS + 11)


if __name__ == '__main__':
  absltest.main()



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Core population utilities."""
import collections
import json
from typing import Any, Callable, Collection, List, Optional, Sequence, Tuple, Type
import attr
import numpy as np
from vizier import pyvizier as vz
from vizier._src.algorithms.evolution import templates
from vizier.interfaces import serializable
from vizier.pyvizier import converters
from vizier.utils import json_utils

# TODO: Use a byte encoding instead of JSON.


def _filter_and_split(
    metrics: Collection[vz.MetricInformation],
) -> Tuple[List[vz.MetricInformation], List[vz.MetricInformation]]:
  """Choose objective and safety metrics and split.

  Args:
    metrics:

  Returns:
    Tuple of objective and safety metrics.
  """
  metrics_by_type = collections.defaultdict(list)
  for metric in metrics:
    metrics_by_type[metric.type].append(metric)

  return (
      metrics_by_type[vz.MetricType.OBJECTIVE],
      metrics_by_type[vz.MetricType.SAFETY],
  )


def _concat(a1: np.ndarray, a2: np.ndarray) -> np.ndarray:
  return np.concatenate([a1, a2], axis=0)


def _shape_equals(
    instance_to_shape: Callable[[Any], Collection[Optional[int]]]
):
  """Creates a shape validator for attr.

  For example, _shape_equals(lambda s : [3, None]) validates that the shape has
  length 2 and its first element is 3.

  Args:
    instance_to_shape: Takes instance as input and returns the desired shape for
      the instance. `None` is treated as "any number".

  Returns:
    A validator that can be passed into attr.ib or attr.field.
  """

  def validator(instance, attribute, value) -> None:
    shape = instance_to_shape(instance)

    def _validator_boolean():
      if len(value.shape) != len(shape):
        return False
      for s1, s2 in zip(value.shape, shape):
        if (s2 is not None) and (s1 != s2):
          return False
      return True

    if not _validator_boolean():
      raise ValueError(
          f'{attribute.name} has shape {value.shape} '
          f'which does not match the expected shape {shape}'
      )

  return validator


@attr.define(frozen=True, init=False)
class Offspring(serializable.Serializable):
  """Offspring which corresponds to trial suggestions.

  Attributes:
    xs: Encoding of trial parameters.
    generations: The number of changes this gene went through.
    ids: Identifier for the gene. Can be used to track the "family tree" created
      from the execution of an evolutionary algorithm. For example, when a new
      offspring is sampled, it is assigned a new id and its successors all carry
      the same id.
  """

  xs: np.ndarray = attr.field(
      validator=[
          attr.validators.instance_of(np.ndarray),
          _shape_equals(lambda s: [len(s), None]),
      ]
  )
  ids: np.ndarray = attr.field(
      validator=[
          attr.validators.instance_of(np.ndarray),
          _shape_equals(lambda s: [len(s)]),
      ]
  )

  generations: np.ndarray = attr.field(
      validator=[
          attr.validators.instance_of(np.ndarray),
          _shape_equals(lambda s: [len(s)]),
      ]
  )

  def __init__(
      self,
      xs: np.ndarray,
      ids: Optional[np.ndarray] = None,
      generations: Optional[np.ndarray] = None,
  ):
    if generations is None:
      generations = np.zeros([xs.shape[0]])
    if ids is None:
      ids = np.zeros([xs.shape[0]])

    self.__attrs_init__(xs, ids, generations)

  def __len__(self) -> int:
    return self.generations.shape[0]

  def __getitem__(self, index: Any) -> 'Offspring':
    return Offspring(self.xs[index], self.ids[index], self.generations[index])

  def __add__(self, other: 'Offspring') -> 'Offspring':
    return Offspring(
        _concat(self.xs, other.xs),
        _concat(self.ids, other.ids),
        _concat(self.generations, other.generations),
    )

  @classmethod
  def load(cls: Type['Offspring'], metadata: vz.Metadata) -> 'Offspring':
    encoded = metadata.get('values', cls=str)
    try:
      decoded = json.loads(encoded, object_hook=json_utils.numpy_hook)
    except Exception as e:
      raise serializable.DecodeError('Failed to decode') from e
    return cls(**decoded)

  def dump(self) -> vz.Metadata:
    encoded = json.dumps(attr.asdict(self), cls=json_utils.NumpyEncoder)
    return vz.Metadata({'values': encoded})


@attr.define(frozen=True)
class Population(templates.Population):
  """Population.

  No validations are done.

  Attributes:
    xs: [len(trials), D] array where D is the number of features. The values are
      in [0, 1] range.
    ys: [len(trials), M_1] array where M_1 is the number of objective metrics.
      The values should be pre-processed as maximization metrics.
    cs: [len(trials), M_2] array where M_2 is the number of soft constraint
      metrics. The values should be pre-processed such that the constraint is to
      get them greater than zero.
    ages: The number of selection phases that each entity survived for.
    generations: The number of ancestors of each entity.
    ids: Identifier for the gene aka "family name".
    trial_ids: Trial ids. For debugging and testing only.
  """

  xs: np.ndarray = attr.field(
      validator=[
          attr.validators.instance_of(np.ndarray),
          _shape_equals(lambda s: [len(s), None]),
      ]
  )
  ys: np.ndarray = attr.field(
      validator=[
          attr.validators.instance_of(np.ndarray),
          _shape_equals(lambda s: [len(s), None]),
      ]
  )
  cs: np.ndarray = attr.field(
      validator=[
          attr.validators.instance_of(np.ndarray),
          _shape_equals(lambda s: [len(s), None]),
      ]
  )
  ages: np.ndarray = attr.field(
      validator=[
          attr.validators.instance_of(np.ndarray),
          _shape_equals(lambda s: [len(s)]),
      ]
  )
  generations: np.ndarray = attr.field(
      validator=[
          attr.validators.instance_of(np.ndarray),
          _shape_equals(lambda s: [len(s)]),
      ]
  )
  ids: np.ndarray = attr.field(
      validator=[
          attr.validators.instance_of(np.ndarray),
          _shape_equals(lambda s: [len(s)]),
      ]
  )
  trial_ids: np.ndarray = attr.field(
      validator=[
          attr.validators.instance_of(np.ndarray),
          _shape_equals(lambda s: [len(s)]),
      ]
  )

  def __len__(self) -> int:
    return self.ys.shape[0]

  def __getitem__(
      self,
      index: Any,
     
  ) -> 'Population':
    return Population(**{k: v[index] for k, v in attr.asdict(self).items()})

  def __add__(self, other: 'Population') -> 'Population':
    return Population(
        _concat(self.xs, other.xs),
        _concat(self.ys, other.ys),
        _concat(self.cs, other.cs),
        _concat(self.ages, other.ages),
        _concat(self.generations, other.generations),
        _concat(self.ids, other.ids),
        _concat(self.trial_ids, other.trial_ids),
    )

  @classmethod
  def recover(cls, metadata: vz.Metadata) -> 'Population':
    encoded = metadata.get('values', default='', cls=str)
    try:
      decoded = json.loads(encoded, object_hook=json_utils.numpy_hook)
    except json.JSONDecodeError as e:
      raise serializable.DecodeError('Failed to recover state.') from e
    return cls(**decoded)

  def dump(self) -> vz.Metadata:
    encoded = json.dumps(attr.asdict(self), cls=json_utils.NumpyEncoder)
    return vz.Metadata({'values': encoded})

  def empty_like(self) -> 'Population':
    """Creates an empty population that has the same shape as this."""

    return Population(
        **{
            k: np.zeros([0] + list(v.shape[1:]))
            for k, v in attr.asdict(self).items()
        }
    )


def _create_parameter_converters(
    search_space: vz.SearchSpace,
) -> Collection[converters.DefaultModelInputConverter]:
  """Returns parameter converters."""
  if search_space.is_conditional:
    raise ValueError('Cannot handle conditional search space!')

  def create_input_converter(
      pc: vz.ParameterConfig,
  ) -> converters.DefaultModelInputConverter:
    return converters.DefaultModelInputConverter(
        pc, scale=True, max_discrete_indices=0, onehot_embed=True
    )

  return [create_input_converter(pc) for pc in search_space.parameters]


def _create_metric_converter(
    mc: vz.MetricInformation,
) -> converters.DefaultModelOutputConverter:
  # TODO: Do something other than raising an error
  return converters.DefaultModelOutputConverter(
      mc,
      flip_sign_for_minimization_metrics=True,
      shift_safe_metrics=True,
      raise_errors_for_missing_metrics=True,
  )


class PopulationConverter(templates.PopulationConverter):
  """Population converter."""

  def __init__(
      self,
      search_space: vz.SearchSpace,
      metrics: Collection[vz.MetricInformation],
      *,
      metadata_ns: str = 'population',
      trial_converter: Optional[converters.DefaultTrialConverter] = None,
  ):
    self._objective_metrics, self._safe_metrics = _filter_and_split(metrics)
    self._num_objective_metrics = len(self._objective_metrics)
    self._num_safe_metrics = len(self._safe_metrics)
    self._metrics = self._objective_metrics + self._safe_metrics
    self._metadata_ns = metadata_ns

    self._trial_converter = trial_converter or converters.DefaultTrialConverter(
        _create_parameter_converters(search_space),
        [_create_metric_converter(mc) for mc in self._metrics],
    )
    self._empty_feature_dict = converters.DictOf2DArrays(
        self._trial_converter.to_features([])
    )

  def to_suggestions(  # pytype: disable=signature-mismatch  # overriding-parameter-type-checks
      self, offsprings: Offspring
  ) -> Collection[vz.TrialSuggestion]:
    parameters_list = self._trial_converter.to_parameters(
        self._empty_feature_dict.dict_like(offsprings.xs)
    )
    suggestions = [vz.TrialSuggestion(p) for p in parameters_list]
    for idx, t in enumerate(suggestions):
      t.metadata.ns(self._metadata_ns).update(offsprings[idx : idx + 1].dump())
    return suggestions

  def _empty_offsprings(self) -> Offspring:
    return Offspring(
        self._empty_feature_dict.asarray(), np.zeros([0]), np.zeros([0])
    )

  def to_population(self, completed: Sequence[vz.CompletedTrial]) -> Population:
    """Converts trials into population. Accepts an empty list."""
    offsprings = self._empty_offsprings()  # create empty
    # Each Trial should contain its genes as metadata. (Note that
    # genes-to-trial mapping is many-to-one). We try to load the genes.
    for t in completed:
      metadata = t.metadata.ns(self._metadata_ns)
      try:
        offsprings += Offspring.load(metadata)
      except serializable.DecodeError:
        # Upon failure, arbitrarily choose one set of genes that map to the
        # current trial.
        offsprings += Offspring(
            converters.DictOf2DArrays(
                self._trial_converter.to_features([t])
            ).asarray(),
            np.zeros([1]),
            np.zeros([1]),
        )

    ys = self._trial_converter.to_labels_array(completed)
    return Population(
        offsprings.xs,
        ys[:, : self._num_objective_metrics],
        ys[:, self._num_objective_metrics :],
        np.zeros([ys.shape[0]]),
        offsprings.generations,
        offsprings.ids,
        np.asarray([t.id for t in completed], dtype=np.int32),
    )


class UniformRandomSampler(templates.Sampler[Offspring]):
  """Generates uniformly random samples."""

  def __init__(self, search_space: vz.SearchSpace, seed: Optional[int] = None):
    if search_space.is_conditional:
      raise ValueError(f'{type(self)} does not support conditional spaces.')
    self._trial_converter = converters.DefaultTrialConverter(
        _create_parameter_converters(search_space)
    )
    self._dimension = sum(
        v[-1] for v in self._trial_converter.features_shape.values()
    )
    self._num_samples = 0
    self._rng = np.random.RandomState(seed)

  def sample(self, count: int) -> Offspring:
    self._num_samples += count
    return Offspring(
        self._rng.random([count, self._dimension]),
        ids=np.arange(self._num_samples - count, self._num_samples),
    )


class LinfMutation(templates.Mutation):
  """L-inf Mutations. Values are assumed to be in [0,1] range."""

  def __init__(self, norm: float = 0.1, seed: Optional[int] = None):
    self._norm = norm
    self._rng = np.random.RandomState(seed)

  def mutate(self, population: Population, count: int) -> Offspring:
    """Perturb by a uniform sample from l-inf ball.

    If the perturbation pushes a coordinate out of [0, 1] range, it "wraps
    around". This is likely suboptimal if optimum is at or near the boundary.

    Args:
      population:
      count:

    Returns:
      Offsprings.
    """
    # Mutate and truncate to [-5, 1.5] range. Perturbations should generally
    # NOT generate values outside this range.
    arr = population.xs.copy()
    arr += self._rng.uniform(-self._norm, self._norm, arr.shape)
    arr = np.maximum(np.minimum(arr, 1.5), -0.5)

    # When the value is outside [0, 1] range, wrap around the limits.
    # -0.2 becomes 0.8, 1.2 becomes 0.2, etc.
    arr = np.where(arr < 0, 1 + arr, np.where(arr > 1, arr - 1, arr))
    return Offspring(arr, population.ids, population.generations + 1)



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Exceptions used in the PythiaServer and Pythia Policy code."""


class PythiaError(Exception):
  """Abstract superclass of Pythia-related errors."""


class TemporaryPythiaError(PythiaError):
  """Vizier should attempt this interaction with a different Pythia task.

  Policy code should raise this to terminate and retry its
  interaction with Vizier.  Maps to an ABORTED gRPC error code.

  Use case: Catch-all errors where there's some hope that a re-run will succeed
    (e.g. OOM, timeouts, snapshot/version problems, or failure of
    nondeterministic algorithms).  If a Policy raises this error, Vizier will
    re-try the computation a modest number of times.
  """


class InactivateStudyError(PythiaError):
  """Pythia cannot handle this Study as configured; inactivate it.

  This terminates the interaction and returns an ARGUMENT_INVALID gRPC error
  code.  It's intended for unrecoverable errors, or misconfigued Studies.
  There will be no re-tries.
  """


class LoadTooLargeError(PythiaError):
  """Raised when the Pythia server is overly busy.

  If a Policy raises this error, it will be retried a (nearly) infinite number
  of times by the Vizier server, on the assumption that there will eventually
  be a server that's not overly busy.
  """


class CancelComputeError(Exception):
  """Raised by the Pythia server to kill Policy computations.

  This is raised in the Pythia compute thread (it is raised by the
  PolicySupporter object) when Vizier cancels the interaction. Policy code
  may catch this temporarily to do clean-up actions, but should re-raise it.

  Vizier may cancel the interaction when the SuggestionWorker needs to shut down
  (i.e. because it's preempted), or when the Pythia computation runs too long.
  """


class PythiaProtocolError(Exception):
  """There was a problem with the interaction between Vizier and Pythia.

  This implies a bug in the protocol code.  Do not raise or catch this in
  Policy code; please report any occurrences to vizier-team@google.com.
  Maps to an INTERNAL gRPC error code.
  """


class VizierDatabaseError(Exception):
  """Vizier had an error on a database operation.

  This can mean an attempt to access unavailable data, access control, etc.
  """



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Simple4d convergence tests for Eagle designer."""

from absl.testing import parameterized
from vizier._src.algorithms.designers.eagle_strategy import eagle_strategy
from vizier._src.algorithms.testing import simplekd_runner
from vizier._src.benchmarks.experimenters.synthetic import simplekd
from absl.testing import absltest


class SimpleKDEagleDesignerTest(parameterized.TestCase):

  @parameterized.parameters(
      dict(best_category='corner'),
      dict(best_category='center'),
      dict(best_category='mixed'),
  )
  def test_simple4d(self, best_category: simplekd.SimpleKDCategory) -> None:
    def _eagle_designer_factory(problem, seed):
      return eagle_strategy.EagleStrategyDesigner(problem, seed=seed)

    simplekd_runner.SimpleKDConvergenceTester(
        best_category=best_category,
        designer_factory=_eagle_designer_factory,
        num_trials=5000,
        max_relative_error=0.05,
        num_repeats=20,
        target_num_convergence=10,
        is_deterministic=True,
    ).assert_convergence()


if __name__ == '__main__':
  absltest.main()



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""PyVizier classes relevant for the service, including OSS-specific objects."""

# TODO: Re-evalaute what to expose to users, once closed.
# TODO: Alternatively, remove this import.
from vizier._src.pyvizier.oss import metadata_util
from vizier._src.pyvizier.oss.automated_stopping import AutomatedStoppingConfig
from vizier._src.pyvizier.oss.automated_stopping import AutomatedStoppingConfigProto
from vizier._src.pyvizier.oss.proto_converters import EarlyStopConverter
from vizier._src.pyvizier.oss.proto_converters import MeasurementConverter
from vizier._src.pyvizier.oss.proto_converters import MetadataDeltaConverter
from vizier._src.pyvizier.oss.proto_converters import MonotypeParameterSequence
from vizier._src.pyvizier.oss.proto_converters import ParameterConfigConverter
from vizier._src.pyvizier.oss.proto_converters import ParameterType
from vizier._src.pyvizier.oss.proto_converters import ParameterValueConverter
from vizier._src.pyvizier.oss.proto_converters import ProblemStatementConverter
from vizier._src.pyvizier.oss.proto_converters import ScaleType
from vizier._src.pyvizier.oss.proto_converters import StudyStateConverter
from vizier._src.pyvizier.oss.proto_converters import SuggestConverter
from vizier._src.pyvizier.oss.proto_converters import TrialConverter
from vizier._src.pyvizier.oss.proto_converters import TrialSuggestionConverter
from vizier._src.pyvizier.oss.study_config import Algorithm
from vizier._src.pyvizier.oss.study_config import ObservationNoise
from vizier._src.pyvizier.oss.study_config import ParameterValueSequence
from vizier._src.pyvizier.oss.study_config import StudyConfig
from vizier._src.pyvizier.pythia.study import StudyDescriptor
from vizier._src.pyvizier.pythia.study import StudyState
from vizier._src.pyvizier.pythia.study import StudyStateInfo
from vizier._src.pyvizier.shared.base_study_config import MetricInformation
from vizier._src.pyvizier.shared.base_study_config import MetricsConfig
from vizier._src.pyvizier.shared.base_study_config import ObjectiveMetricGoal
from vizier._src.pyvizier.shared.base_study_config import ProblemStatement
from vizier._src.pyvizier.shared.common import Metadata
from vizier._src.pyvizier.shared.common import MetadataValue
from vizier._src.pyvizier.shared.common import Namespace
from vizier._src.pyvizier.shared.parameter_config import ExternalType
from vizier._src.pyvizier.shared.parameter_config import ParameterConfig
from vizier._src.pyvizier.shared.parameter_config import SearchSpace
from vizier._src.pyvizier.shared.parameter_config import SearchSpaceSelector
from vizier._src.pyvizier.shared.study import ProblemAndTrials
from vizier._src.pyvizier.shared.trial import CompletedTrial
from vizier._src.pyvizier.shared.trial import CompletedTrialWithMeasurements
from vizier._src.pyvizier.shared.trial import Measurement
from vizier._src.pyvizier.shared.trial import MetadataDelta
from vizier._src.pyvizier.shared.trial import Metric
from vizier._src.pyvizier.shared.trial import NaNMetric
from vizier._src.pyvizier.shared.trial import ParameterDict
from vizier._src.pyvizier.shared.trial import ParameterValue
from vizier._src.pyvizier.shared.trial import PendingTrial
from vizier._src.pyvizier.shared.trial import PendingTrialWithMeasurements
from vizier._src.pyvizier.shared.trial import Trial
from vizier._src.pyvizier.shared.trial import TrialFilter
from vizier._src.pyvizier.shared.trial import TrialStatus
from vizier._src.pyvizier.shared.trial import TrialSuggestion



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Basic datastore written using Python primitives.

For debugging/testing purposes mainly.
"""
import collections
import copy
import dataclasses
import threading
from typing import Callable, DefaultDict, Dict, Iterable, List, Optional
from absl import logging

from vizier._src.service import custom_errors
from vizier._src.service import datastore
from vizier._src.service import key_value_pb2
from vizier._src.service import resources
from vizier._src.service import study_pb2
from vizier._src.service import vizier_oss_pb2
from vizier._src.service import vizier_service_pb2
from vizier.service import pyvizier as vz

from google.longrunning import operations_pb2

UnitMetadataUpdate = vizier_service_pb2.UnitMetadataUpdate


@dataclasses.dataclass(frozen=True)
class ClientNode:
  """Only contains suggestion operations associated with this client."""

  # Keys are `operation_id`.
  suggestion_operations: Dict[str, operations_pb2.Operation] = (
      dataclasses.field(default_factory=dict)
  )


# Specific dataclasses used for NestedDictRAMDataStore.
@dataclasses.dataclass(frozen=True)
class StudyNode:
  """Contains original study and currently stored trials."""

  study_proto: study_pb2.Study

  # Keys are `trial_id`.
  trial_protos: Dict[int, study_pb2.Trial] = dataclasses.field(
      default_factory=dict
  )

  # Keys are `operation_id`.
  early_stopping_operations: Dict[
      str, vizier_oss_pb2.EarlyStoppingOperation
  ] = dataclasses.field(default_factory=dict)

  # Key is `client_id`, which distinguishes clients, so you can have several
  # clients interacting with the same Vizier server.
  clients: Dict[str, ClientNode] = dataclasses.field(default_factory=dict)


@dataclasses.dataclass(frozen=True)
class OwnerNode:
  """First level of the entire RAM Datastore."""

  # Key is a `study_id` that pick's out one of the owner's Studies.
  studies: Dict[str, StudyNode] = dataclasses.field(default_factory=dict)


class NestedDictRAMDataStore(datastore.DataStore):
  """Basic Datastore class using nested dictionaries."""

  def __init__(self):
    """Organization as follows."""
    # Key is `owner_id`, which corresponds to the (perhaps human) owner
    # of the Study.
    self._owners: Dict[str, OwnerNode] = {}
    # TODO: Use more fine-grained (study/trial/etc.) locks.
    self._lock = threading.Lock()

  def create_study(self, study: study_pb2.Study) -> resources.StudyResource:
    resource = resources.StudyResource.from_name(study.name)
    temp_dict = {resource.study_id: StudyNode(study_proto=copy.deepcopy(study))}

    with self._lock:
      if resource.owner_id not in self._owners:
        self._owners[resource.owner_id] = OwnerNode(studies=temp_dict)
      else:
        study_dict = self._owners[resource.owner_id].studies
        if resource.study_id not in study_dict:
          study_dict.update(temp_dict)
        else:
          raise custom_errors.AlreadyExistsError(
              'Study with that name already exists.', study.name
          )
    return resource

  def load_study(self, study_name: str) -> study_pb2.Study:
    resource = resources.StudyResource.from_name(study_name)
    try:
      with self._lock:
        return copy.deepcopy(
            self._owners[resource.owner_id]
            .studies[resource.study_id]
            .study_proto
        )
    except KeyError as err:
      raise custom_errors.NotFoundError(
          'Could not get Study with name:', resource.name
      ) from err

  def update_study(self, study: study_pb2.Study) -> resources.StudyResource:
    resource = resources.StudyResource.from_name(study.name)
    try:
      with self._lock:
        self._owners[resource.owner_id].studies[
            resource.study_id
        ].study_proto.CopyFrom(study)
      return resource
    except KeyError as err:
      raise custom_errors.NotFoundError(
          'Could not update Study with name:', resource.name
      ) from err

  def delete_study(self, study_name: str) -> None:
    resource = resources.StudyResource.from_name(study_name)
    try:
      with self._lock:
        del self._owners[resource.owner_id].studies[resource.study_id]
    except KeyError as err:
      raise custom_errors.NotFoundError(
          'Study does not exist:', study_name
      ) from err

  def list_studies(self, owner_name: str) -> List[study_pb2.Study]:
    resource = resources.OwnerResource.from_name(owner_name)
    try:
      with self._lock:
        study_nodes = list(self._owners[resource.owner_id].studies.values())
        return copy.deepcopy(
            [study_node.study_proto for study_node in study_nodes]
        )
    except KeyError as err:
      raise custom_errors.NotFoundError(
          'Owner does not exist:', owner_name
      ) from err

  def create_trial(self, trial: study_pb2.Trial) -> resources.TrialResource:
    resource = resources.TrialResource.from_name(trial.name)
    with self._lock:
      trial_protos = (
          self._owners[resource.owner_id]
          .studies[resource.study_id]
          .trial_protos
      )
      if resource.trial_id in trial_protos:
        raise custom_errors.AlreadyExistsError(
            'Trial %s already exists' % trial.name
        )
      else:
        trial_protos[resource.trial_id] = copy.deepcopy(trial)
    return resource

  def get_trial(self, trial_name: str) -> study_pb2.Trial:
    resource = resources.TrialResource.from_name(trial_name)
    try:
      with self._lock:
        return copy.deepcopy(
            self._owners[resource.owner_id]
            .studies[resource.study_id]
            .trial_protos[resource.trial_id]
        )
    except KeyError as err:
      raise custom_errors.NotFoundError(
          'Could not get Trial with name:', resource.name
      ) from err

  def update_trial(self, trial: study_pb2.Trial) -> resources.TrialResource:
    resource = resources.TrialResource.from_name(trial.name)
    try:
      with self._lock:
        trial_protos = (
            self._owners[resource.owner_id]
            .studies[resource.study_id]
            .trial_protos
        )
        if resource.trial_id not in trial_protos:
          raise custom_errors.NotFoundError(
              'Trial %s does not exist.' % trial.name
          )
        trial_protos[resource.trial_id] = copy.deepcopy(trial)
      return resource
    except KeyError as err:
      raise custom_errors.NotFoundError(
          'Could not update Trial with name:', resource.name
      ) from err

  def list_trials(self, study_name: str) -> List[study_pb2.Trial]:
    resource = resources.StudyResource.from_name(study_name)
    try:
      with self._lock:
        return copy.deepcopy(
            list(
                self._owners[resource.owner_id]
                .studies[resource.study_id]
                .trial_protos.values()
            )
        )
    except KeyError as err:
      raise custom_errors.NotFoundError(
          'Study does not exist:', study_name
      ) from err

  def delete_trial(self, trial_name: str) -> None:
    resource = resources.TrialResource.from_name(trial_name)
    try:
      with self._lock:
        del (
            self._owners[resource.owner_id]
            .studies[resource.study_id]
            .trial_protos[resource.trial_id]
        )
    except KeyError as err:
      raise custom_errors.NotFoundError(
          'Trial does not exist:', trial_name
      ) from err

  def max_trial_id(self, study_name: str) -> int:
    resource = resources.StudyResource.from_name(study_name)
    try:
      with self._lock:
        trial_ids = copy.deepcopy(
            list(
                self._owners[resource.owner_id]
                .studies[resource.study_id]
                .trial_protos.keys()
            )
        )
    except KeyError as err:
      raise custom_errors.NotFoundError(
          'Study does not exist:', study_name
      ) from err

    if trial_ids:
      return max(trial_ids)
    else:
      return 0

  def create_suggestion_operation(
      self, operation: operations_pb2.Operation
  ) -> resources.SuggestionOperationResource:
    resource = resources.SuggestionOperationResource.from_name(operation.name)
    with self._lock:
      if (
          resource.client_id
          not in self._owners[resource.owner_id]
          .studies[resource.study_id]
          .clients
      ):
        self._owners[resource.owner_id].studies[resource.study_id].clients[
            resource.client_id
        ] = ClientNode()
      suggestion_operations = (
          self._owners[resource.owner_id]
          .studies[resource.study_id]
          .clients[resource.client_id]
          .suggestion_operations
      )

      if resource.operation_id in suggestion_operations:
        raise custom_errors.AlreadyExistsError(
            'Operation already exists:', resource.operation_id
        )

      suggestion_operations[resource.operation_id] = copy.deepcopy(operation)
    return resource

  def get_suggestion_operation(
      self, operation_name: str
  ) -> operations_pb2.Operation:
    resource = resources.SuggestionOperationResource.from_name(operation_name)
    try:
      with self._lock:
        return copy.deepcopy(
            self._owners[resource.owner_id]
            .studies[resource.study_id]
            .clients[resource.client_id]
            .suggestion_operations[resource.operation_id]
        )

    except KeyError as err:
      raise custom_errors.NotFoundError(
          'Could not find SuggestionOperation with name:', resource.name
      ) from err

  def update_suggestion_operation(
      self, operation: operations_pb2.Operation
  ) -> resources.SuggestionOperationResource:
    resource = resources.SuggestionOperationResource.from_name(operation.name)
    try:
      with self._lock:
        self._owners[resource.owner_id].studies[resource.study_id].clients[
            resource.client_id
        ].suggestion_operations[resource.operation_id] = copy.deepcopy(
            operation
        )
      return resource
    except KeyError as err:
      raise custom_errors.NotFoundError(
          'Could not update SuggestionOperation with name:', resource.name
      ) from err

  def list_suggestion_operations(
      self,
      study_name: str,
      client_id: str,
      filter_fn: Optional[Callable[[operations_pb2.Operation], bool]] = None,
  ) -> List[operations_pb2.Operation]:
    resource = resources.StudyResource.from_name(study_name)
    try:
      with self._lock:
        operations_list = copy.deepcopy(
            list(
                self._owners[resource.owner_id]
                .studies[resource.study_id]
                .clients[client_id]
                .suggestion_operations.values()
            )
        )
    except KeyError as err:
      raise custom_errors.NotFoundError(
          '(study_name, client_id) does not exist:', (study_name, client_id)
      ) from err

    if filter_fn is not None:
      return copy.deepcopy([op for op in operations_list if filter_fn(op)])
    else:
      return copy.deepcopy(operations_list)

  def max_suggestion_operation_number(
      self, study_name: str, client_id: str
  ) -> int:
    resource = resources.StudyResource.from_name(study_name)
    try:
      with self._lock:
        ops = (
            self._owners[resource.owner_id]
            .studies[resource.study_id]
            .clients[client_id]
            .suggestion_operations
        )
        return len(ops)
    except KeyError as err:
      raise custom_errors.NotFoundError(
          '(study_name, client_id) does not exist:', (study_name, client_id)
      ) from err

  def create_early_stopping_operation(
      self, operation: vizier_oss_pb2.EarlyStoppingOperation
  ) -> resources.EarlyStoppingOperationResource:
    resource = resources.EarlyStoppingOperationResource.from_name(
        operation.name
    )
    with self._lock:
      early_stopping_ops = (
          self._owners[resource.owner_id]
          .studies[resource.study_id]
          .early_stopping_operations
      )
      if resource.operation_id in early_stopping_ops:
        raise custom_errors.AlreadyExistsError(
            'Operation already exists:', resource.operation_id
        )

      early_stopping_ops[resource.operation_id] = copy.deepcopy(operation)
    return resource

  def get_early_stopping_operation(
      self, operation_name: str
  ) -> vizier_oss_pb2.EarlyStoppingOperation:
    resource = resources.EarlyStoppingOperationResource.from_name(
        operation_name
    )
    try:
      with self._lock:
        return copy.deepcopy(
            self._owners[resource.owner_id]
            .studies[resource.study_id]
            .early_stopping_operations[resource.operation_id]
        )
    except KeyError as err:
      raise custom_errors.NotFoundError(
          'Could not find EarlyStoppingOperation with name:', resource.name
      ) from err

  def update_early_stopping_operation(
      self, operation: vizier_oss_pb2.EarlyStoppingOperation
  ) -> resources.EarlyStoppingOperationResource:
    resource = resources.EarlyStoppingOperationResource.from_name(
        operation.name
    )
    try:
      with self._lock:
        self._owners[resource.owner_id].studies[
            resource.study_id
        ].early_stopping_operations[resource.operation_id] = copy.deepcopy(
            operation
        )
      return resource
    except KeyError as err:
      raise custom_errors.NotFoundError(
          'Could not update EarlyStoppingOperation with name:', resource.name
      ) from err

  def update_metadata(
      self,
      study_name: str,
      study_metadata: Iterable[key_value_pb2.KeyValue],
      trial_metadata: Iterable[UnitMetadataUpdate],
  ) -> None:
    # TODO:
    """Writes the supplied metadata to the database.

    Args:
      study_name:
      study_metadata: Metadata that's associated with the Study as a whole.
      trial_metadata: Metadata that's associated with trials.  (Note that its an
        error to attach metadata to a Trial that doesn't exist.)
    """
    s_resource = resources.StudyResource.from_name(study_name)
    logging.debug('database.update_metadata s_resource= %s', s_resource)

    with self._lock:
      try:
        study_node = self._owners[s_resource.owner_id].studies[
            s_resource.study_id
        ]
      except KeyError as e:
        raise custom_errors.NotFoundError(
            'No such study:', s_resource.name
        ) from e
      # Store Study-related metadata into the database.
      vz.metadata_util.merge_study_metadata(
          study_node.study_proto.study_spec, copy.deepcopy(study_metadata)
      )
      # Split the trial-related metadata by Trial.
      split_metadata: DefaultDict[str, List[UnitMetadataUpdate]] = (
          collections.defaultdict(list)
      )
      for md in copy.deepcopy(trial_metadata):
        split_metadata[md.trial_id].append(md)
      # Now, we update one Trial at a time:
      for trial_id, md_list in split_metadata.items():
        t_resource = s_resource.trial_resource(trial_id)
        trial_proto = study_node.trial_protos[t_resource.trial_id]
        vz.metadata_util.merge_trial_metadata(trial_proto, md_list)



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Tests for surrogate_experimenter."""
from typing import Sequence, Optional
import jax

from vizier import algorithms as vza
from vizier import pyvizier as vz
from vizier._src.algorithms.designers import quasi_random
from vizier._src.benchmarks.experimenters import surrogate_experimenter
from vizier.testing import test_studies

from absl.testing import absltest


class DummyPredictor(vza.Predictor):

  def predict(
      self,
      trials: Sequence[vz.TrialSuggestion],
      rng: Optional[jax.Array] = None,
      num_samples: Optional[int] = None,
  ) -> vza.Prediction:
    num_trials = len(trials)
    mean = jax.random.normal(key=rng, shape=(num_trials,))
    stddev = jax.random.normal(key=rng, shape=(num_trials,))
    return vza.Prediction(mean=mean, stddev=stddev)


class PredictorExperimenterTest(absltest.TestCase):

  def test_e2e(self):
    problem = vz.ProblemStatement(test_studies.flat_space_with_all_types())
    problem.metric_information.append(
        vz.MetricInformation(
            name='metric', goal=vz.ObjectiveMetricGoal.MAXIMIZE
        )
    )
    predictor = DummyPredictor()
    experimenter = surrogate_experimenter.PredictorExperimenter(
        predictor=predictor, problem_statement=problem
    )

    quasi_random_sampler = quasi_random.QuasiRandomDesigner(
        problem.search_space
    )
    suggestions = quasi_random_sampler.suggest(count=7)
    trials = [suggestion.to_trial() for suggestion in suggestions]
    experimenter.evaluate(trials)

    for trial in trials:
      self.assertEqual(trial.status, vz.TrialStatus.COMPLETED)
      self.assertContainsSubset(
          trial.final_measurement.metrics.keys(), ['metric']
      )


if __name__ == '__main__':
  absltest.main()



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Contains resource utilities, such as name parsing.

Convention for variable naming for string identifiers:
  1. {}_id means single component (e.g. study_id = '1312931').
  2. {}_name means directory (e.g. study_name =
  'owners/my_username/studies/1312931').

Everything related to naming should be self-contained in this file; i.e.
dependents should only call official API functions from this file and never
explicitly write their own string processing.
"""
import re
import attr

from vizier.utils import attrs_utils

# Resource components cannot contain "/".
_resource_component_validator = [attrs_utils.assert_re_fullmatch(r'[^\/]+')]


@attr.define(init=True, frozen=True)
class OwnerResource:
  """Resource for Owners."""

  _owner_id: str = attr.ib(init=True, validator=_resource_component_validator)

  @classmethod
  def from_name(cls, resource_name: str):
    owner_match = re.match(r'^owners\/(?P<owner_id>[^\/]+)$', resource_name)

    if owner_match:
      return OwnerResource(owner_id=owner_match.group('owner_id'))
    else:
      raise ValueError(f'Incorrect resource name sent: {resource_name}')

  @property
  def owner_id(self):
    return self._owner_id

  @property
  def name(self) -> str:
    return f'owners/{self._owner_id}'


@attr.define(init=True, frozen=True)
class StudyResource:
  """Resource for Studies."""

  owner_id: str = attr.ib(init=True, validator=_resource_component_validator)
  study_id: str = attr.ib(init=True, validator=_resource_component_validator)

  @classmethod
  def from_name(cls, resource_name: str):
    """Creates StudyResource from a name."""
    study_match = re.match(
        r'^owners\/(?P<owner_id>[^\/]+)\/studies\/(?P<study_id>[^\/]+)$',
        resource_name,
    )

    if study_match:
      return StudyResource(
          study_match.group('owner_id'), study_match.group('study_id')
      )
    else:
      raise ValueError(
          f'{repr(resource_name)} is not a valid name for a Study resource.'
      )

  @property
  def owner_resource(self) -> OwnerResource:
    return OwnerResource(owner_id=self.owner_id)

  @property
  def name(self) -> str:
    return f'owners/{self.owner_id}/studies/{self.study_id}'

  def trial_resource(self, trial_id: str) -> 'TrialResource':
    """Creates a TrialResource when given a trial_id."""
    int_id = int(trial_id)
    if int_id <= 0:
      raise ValueError('Invalid trial_id: "{trial_id}"')
    return TrialResource(self.owner_id, self.study_id, int_id)


@attr.define(init=True, frozen=True)
class TrialResource:
  """Resource for Trials."""

  owner_id: str = attr.ib(init=True, validator=_resource_component_validator)
  study_id: str = attr.ib(init=True, validator=_resource_component_validator)
  trial_id: int = attr.ib(
      init=True,
      validator=[
          attr.validators.instance_of(int),
          attrs_utils.assert_not_negative,
      ],
  )

  @classmethod
  def from_name(cls, resource_name: str) -> 'TrialResource':
    """Creates TrialResource from a name."""
    trial_match = re.match(
        r'^owners\/(?P<owner_id>[^\/]+)\/studies\/(?P<study_id>[^\/]+)\/trials\/(?P<trial_id>[^\/]+)$',
        resource_name,
    )

    if trial_match:
      return TrialResource(
          trial_match.group('owner_id'),
          trial_match.group('study_id'),
          int(trial_match.group('trial_id')),
      )
    else:
      raise ValueError(
          f'{repr(resource_name)} is not a valid name for a Trial resource.'
      )

  @property
  def study_resource(self) -> StudyResource:
    return StudyResource(owner_id=self.owner_id, study_id=self.study_id)

  @property
  def early_stopping_operation_resource(
      self,
  ) -> 'EarlyStoppingOperationResource':
    return EarlyStoppingOperationResource(
        owner_id=self.owner_id, study_id=self.study_id, trial_id=self.trial_id
    )

  @property
  def name(self) -> str:
    return (
        f'owners/{self.owner_id}/studies/{self.study_id}/trials/{self.trial_id}'
    )


@attr.define(init=True, frozen=True)
class EarlyStoppingOperationResource:
  """Resource for Early Stopping Operations."""

  owner_id: str = attr.ib(init=True, validator=_resource_component_validator)
  study_id: str = attr.ib(init=True, validator=_resource_component_validator)
  trial_id: int = attr.ib(
      init=True,
      validator=[
          attr.validators.instance_of(int),
          attrs_utils.assert_not_negative,
      ],
  )

  @property
  def operation_id(self) -> str:
    return f'earlystopping/{self.study_id}/{self.trial_id}'

  @property
  def name(self) -> str:
    return f'owners/{self.owner_id}/operations/{self.operation_id}'

  @classmethod
  def from_name(cls, resource_name: str):
    """Creates EarlyStoppingOperationResource from a name."""
    operation_match = re.match(
        r'^owners\/(?P<owner_id>[^\/]+)/operations/earlystopping/(?P<study_id>[^\/]+)/(?P<trial_id>[^\/]+)$',
        resource_name,
    )
    if operation_match:
      return EarlyStoppingOperationResource(
          operation_match.group('owner_id'),
          operation_match.group('study_id'),
          int(operation_match.group('trial_id')),
      )
    else:
      raise ValueError(f'Incorrect resource name sent: {resource_name}')

  @property
  def trial_resource(self) -> TrialResource:
    return TrialResource(
        owner_id=self.owner_id, study_id=self.study_id, trial_id=self.trial_id
    )


@attr.define(init=True, frozen=True)
class SuggestionOperationResource:
  """Resource for Suggestion Operations."""

  owner_id: str = attr.ib(init=True, validator=_resource_component_validator)
  study_id: str = attr.ib(init=True, validator=_resource_component_validator)
  client_id: str = attr.ib(init=True, validator=_resource_component_validator)
  operation_number: int = attr.ib(
      init=True,
      validator=[
          attr.validators.instance_of(int),
          attrs_utils.assert_not_negative,
      ],
  )

  @property
  def operation_id(self) -> str:
    return (
        f'suggestion/{self.study_id}/{self.client_id}/{self.operation_number}'
    )

  @property
  def name(self) -> str:
    return f'owners/{self.owner_id}/operations/{self.operation_id}'

  @classmethod
  def from_name(cls, resource_name: str):
    """Creates SuggestionOperationResource from a name."""
    operation_match = re.match(
        r'^owners\/(?P<owner_id>[^\/]+)/operations/suggestion/(?P<study_id>[^\/]+)/(?P<client_id>[^\/]+)/(?P<operation_number>[^\/]+)$',
        resource_name,
    )
    if operation_match:
      return SuggestionOperationResource(
          operation_match.group('owner_id'),
          operation_match.group('study_id'),
          operation_match.group('client_id'),
          int(operation_match.group('operation_number')),
      )
    else:
      raise ValueError(f'Incorrect resource name sent: {resource_name}')



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Meta Learning Designer.

The 'meta-learning' designer attempts to find the optimal hyper-parameters
values of the 'tuned' designer through a meta-learning process. In essensce, it
searches for the hyper-parameters values that yield the best result for the
given problem as oppose to blindly using pre-defined fixed values.

Notes
-----

1) The 'meta-learning' designer's search space is the hyper-parameters of
the 'tuned' designer which we seek to tune.

2) Before tuning starts, a tuned designer instantiated with the default
parameter values defined by the search space is used. This means that
default configuration defined in the designer level will be overridden.

3) The 'tuned_designer_factory' should accept the hyper-parameters as arguments.
This means for example, that if the 'tuned_designer' relies on an internal
configuration class the 'tuned_designer_factory' function would have to handle
the instantiation of that class (see 'eagle_meta_learning_convergence_test.py'
for an example).

4) Each instance of 'tuned' designer is updated with all trials seen thus-far,
therefore hyper-parameters that were created later in the process will benefit
from being instantiated with larger trajectory. This violates the assumption
of the meta-learning designer which the meta-trial metrics are derived from the
same objective function. For now we don't address this issue directly, though
in the future we could consider applying techniques such as RegEvo.
"""

import enum
from typing import Optional, Sequence
import attrs
import numpy as np
from vizier import algorithms as vza
from vizier import pyvizier as vz
from vizier._src.algorithms.designers.meta_learning import meta_learning_utils as utils


@attrs.define
class MetaLearningConfig:
  """Meta learning configuration (with pre-defined default values).

  To balance exploration and exploitation in the meta-learning process it's
  recommended to stop the meta-learner after a certain number of iterations so
  to take advantage of the thus-far best hyper-parameters and not to waste
  suggestion trials on further exploration.

  With the default configuration the number of meta-learning iterations is
  (10000-3000) / 100 = 70. In order to not terminate the meta-learning process,
  set the `tuning_max_num_trials` to sufficiently large value (e.g. 1e6).
  """

  # Number of trials to use per tuning iteration.
  num_trials_per_tuning: int = 100

  # Tuning starts when number of completed trials is at least this threshold.
  tuning_min_num_trials: int = 3000

  # Once the number of trials exceeds this threshold meta-learning stops.
  tuning_max_num_trials: int = 10000


class MetaLearningState(enum.Enum):
  """Meta learning state."""

  # The meta-learning process hasn't started yet, accumulating a sufficient
  # number of trials.
  INITIALIZE = 1

  # The meta-learning progress is performed to tune the designer.
  TUNE = 2

  # A tuned designer with the best hyper-params is used to generate suggestons.
  # Instantiate a tuned designer
  USE_BEST_PARAMS = 3


# TODO: support serialization.
@attrs.define
class MetaLearningDesigner(vza.Designer):
  """Meta learning designer."""

  # The problem associated with the 'tuned' designer.
  problem: vz.ProblemStatement

  # Factory of the 'tuned' designer.
  tuned_designer_factory: vza.DesignerFactory[vza.Designer]

  # Factory of the 'meta-learning' designer.
  meta_designer_factory: vza.DesignerFactory[vza.Designer]

  # The 'tuned' designer hyper-parameters to apply meta-learning on (which
  # constitute the search space of the 'meta-learning' designer).
  tuning_hyperparams: vz.SearchSpace

  # The meta-learner configuration.
  config: Optional[MetaLearningConfig] = None

  # A random seed used for instantiating the designers.
  seed: Optional[int] = None

  # ----------------------------------------------------------------------------
  # Internal Attributes
  # ----------------------------------------------------------------------------

  # The metric name used by the meta-learning designer.
  _meta_designer_metric_name: str = attrs.field(default='score', init=False)

  # The current 'tuned' designer instance with assigned hyper-parameter.
  _curr_tuned_designer: vza.Designer = attrs.field(init=False)

  # The 'meta' designer instance (only instantiated once).
  _meta_designer: vza.Designer = attrs.field(init=False)

  # The current state of the meta-learner.
  _state: MetaLearningState = attrs.field(init=False)

  # All the completed meta-trials suggested by the 'meta' designer.
  _meta_trials: list[vz.Trial] = attrs.field(factory=list, init=False)

  # The current 'tuned' hyper-parameter suggested by the meta-learner designer.
  _curr_tuned_hyperparams: vz.TrialSuggestion = attrs.field(init=False)

  # All the completed trials, suggested by the 'tuned' designers, seen since the
  # beginning of the meta-learner run.
  _trials: list[vz.Trial] = attrs.field(factory=list, init=False)

  # The completed trials suggested by the current 'tuned' designer which was
  # instantiated with the current hyper-parameter values.
  _curr_trials: list[vz.Trial] = attrs.field(factory=list, init=False)

  def __attrs_post_init__(self):
    """Initializes the meta learning desiger."""
    if len(self.problem.metric_information) != 1:
      raise ValueError(f'Expected exactly one metric. {self.problem}')

    if self.seed is None:
      # JAX random seed doesn't accept None, so generating random integer.
      self.seed = np.random.randint(low=0, high=1e6)

    # Instantiate an MetaLearningUtils.
    self._utils = utils.MetaLearningUtils(
        goal=self.problem.metric_information.item().goal,
        tuned_metric_name=self.problem.metric_information.item().name,
        meta_metric_name=self._meta_designer_metric_name,
        tuning_params=self.tuning_hyperparams,
    )
    # Instantiated 'tuned' designer the with default hyper-parameters.
    self._curr_tuned_hyperparams = self._utils.get_default_hyperparameters()
    self._curr_tuned_designer = self.tuned_designer_factory(
        self.problem,
        seed=self.seed,
        **self._curr_tuned_hyperparams.parameters.as_dict(),
    )
    self._meta_designer = self.meta_designer_factory(
        self._utils.meta_problem, seed=self.seed
    )
    self.config = self.config or MetaLearningConfig()
    self._state = MetaLearningState.INITIALIZE

  def suggest(self, count: int = 1) -> Sequence[vz.TrialSuggestion]:
    """Suggests trials."""
    return self._curr_tuned_designer.suggest(count)

  def update(
      self, completed: vza.CompletedTrials, all_active: vza.ActiveTrials
  ) -> None:
    """Incorporates completed trials into the meta-learner designer's state.

    1) At the beginning of the run the state is INITIALIZE, during which a
    'tune' designer with default hyper-parameter is used to suggest trials.

    2) After accumulating `tuning_min_num_trials` completed trials the state
    transitions to TUNE, and then the tuning process starts.

    3) After accumulating `num_trials_per_tuning` completd trials, the meta
    iteration is summarized, during which the best trial is selected to be used
    to update the 'meta' desinger.

    4) Once reaching maximum number of total completed meta-learn trials
    (`tuning_max_num_trials`), the meta-learning process is finalized during
    which the best hyper-parameter values are selected to instantiate a new
    'tuned' designer which is then updated with all seen trials.

    Args:
      completed:
      all_active:
    """
    self._trials.extend(completed.trials)
    self._curr_trials.extend(completed.trials)
    # Update curr_designer with newly completed trials (applies to all states).
    self._curr_tuned_designer.update(
        vza.CompletedTrials(completed.trials), vza.ActiveTrials()
    )

    # Check if enough trials already accumulated to start meta learning.
    if len(self._trials) < self.config.tuning_min_num_trials:
      return

    # Check if the meta-learning process should be terminated.
    elif len(self._trials) >= self.config.tuning_max_num_trials:
      # Check if the meta-learning has just terminated. If so, finalize it.
      if self._state == MetaLearningState.TUNE:
        # Find the best meta-learn result.
        best_meta_trial = self._utils.get_best_meta_trial(self._meta_trials)
        self._curr_tuned_designer = self.tuned_designer_factory(
            self.problem, seed=self.seed, **best_meta_trial.parameters.as_dict()
        )
        # Update the newly created designer with all completed trials.
        self._curr_tuned_designer.update(
            vza.CompletedTrials(self._trials), vza.ActiveTrials()
        )
        self._state = MetaLearningState.USE_BEST_PARAMS

    else:
      self._state = MetaLearningState.TUNE
      # Check if there's enough trials to summarize meta iteration.
      if len(self._curr_trials) >= self.config.num_trials_per_tuning:
        # Get best score for the current iteration.
        meta_trial = self._utils.complete_meta_suggestion(
            meta_suggestion=self._curr_tuned_hyperparams,
            score=self._utils.get_best_tuned_trial_score(self._curr_trials),
        )
        self._meta_designer.update(
            vza.CompletedTrials([meta_trial]), vza.ActiveTrials()
        )
        # Store the iteration results to be used during meta-learn finalization.
        self._meta_trials.append(meta_trial)
        # Get new tuned params suggestion and initialize a new curr_designer.
        self._curr_tuned_hyperparams = self._meta_designer.suggest(1)[0]
        self._curr_tuned_designer = self.tuned_designer_factory(
            self.problem,
            seed=self.seed,
            **self._curr_tuned_hyperparams.parameters.as_dict(),
        )
        # Update the newly created designer with all completed trials.
        self._curr_tuned_designer.update(
            vza.CompletedTrials(self._trials), vza.ActiveTrials()
        )
        # Reset the trials associated with current hyper-parameters.
        self._curr_trials = []



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""The Policy can use these classes to communicate with Vizier."""

import abc
import datetime
from typing import Iterable, List, Optional

from vizier import pyvizier as vz


class PolicySupporter(abc.ABC):
  """Used by Policy instances to communicate with Vizier."""

  # TODO: Change to GetStudyDescriptor (Maybe: Note that
  #   StudyDescriptor includes $max_trial_id, so it causes many more database
  #   collisions than ProblemStatement, especially if there are lots of
  #   workers.)
  @abc.abstractmethod
  def GetStudyConfig(self, study_guid: str) -> vz.ProblemStatement:
    """Requests a StudyConfig from Vizier.

    This sends a PythiaToVizier.trial_selector packet and waits for the
    response(s).  You can call this multiple times, and it is thread-friendly,
    so you can even overlap calls.

    Args:
      study_guid: The GUID of the study whose StudyConfig you want. Note that
        access control applies.

    Returns:
      The requested StudyConfig proto.

    Raises:
      CancelComputeError: (Do not catch.)
      PythiaProtocolError: (Do not catch.)
      VizierDatabaseError: If the database operation raises an error, e.g. if
        $study_guid refers to a nonexistent or inaccessible study.
    """

  # TODO: Should take `TrialFilter` as input, instead of
  # its fields listed as keyword arguments.
  @abc.abstractmethod
  def GetTrials(
      self,
      *,
      study_guid: Optional[str] = None,
      trial_ids: Optional[Iterable[int]] = None,
      min_trial_id: Optional[int] = None,
      max_trial_id: Optional[int] = None,
      status_matches: Optional[vz.TrialStatus] = None,
      include_intermediate_measurements: bool = True
  ) -> List[vz.Trial]:
    """Requests Trials from Vizier.

    Args:
      study_guid: The GUID of the study to get Trials from. If None, uses the
        current Study.
      trial_ids: a list of Trial id numbers to acquire (if None, allows all
        Trials.)
      min_trial_id: Trials with Trial.id >= min_trial ID are selected, if not
        None.
      max_trial_id: Trials with Trial.id <= min_trial ID are selected, if not
        None.
      status_matches: If not None, filters for Trials where
        Trial.status==status_matches.  Selects all Trials by default.
      include_intermediate_measurements: If True (default), the returned Trials
        will include all intermediate measurements.  If False, PolicySupporter
        _may_ leave the `measurements` field empty in the returned Trials (e.g.
        to optimize speed).

    Note that the $final_measurement field will always be included when
      available, i.e. for COMPLETED Trials.

    Returns:
      Trials obtained from Vizier, in order of increasing Trial ID.
      Each argument selects a subset of the Study's Trials, and the result is
      the intersection of the subsets.

    Raises:
      CancelComputeError: (Do not catch.)
      PythiaProtocolError: (Do not catch.)
      VizierDatabaseError: If the database operation raises an error, e.g. if
        $study_guid refers to a nonexistent or inaccessible study.
    """

  @property
  @abc.abstractmethod
  def study_guid(self) -> str:
    """Default study GUID."""

  def CheckCancelled(self, note: Optional[str] = None) -> None:
    """Throws a CancelComputeError on timeout or if Vizier cancels.

    This should be called occasionally by any long-running computation.
    Raises an exception if the interaction has been cancelled by the Vizier
    side of the protocol; the exception shuts down the Pythia server.

    Args:
      note: for debugging.

    Raises:
      CancelComputeError: (Do not catch.)
    """
    pass

  def TimeRemaining(self) -> datetime.timedelta:
    """The time remaining to compute a result.

    Returns:
      The remaining time before the RPC is considered to have timed out; it
      returns datetime.timedelta.max if no deadline was specified in the RPC.

    This is an alternative to calling CheckCancelled(); both have the goal of
    terminating runaway computations.  If your computation times out,
    you should raise TemporaryPythiaError (if you want a retry) or
    InactivateStudyError (if not).
    """
    return datetime.timedelta(hours=1.0)



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Tools for iterating through a collection of Parameter(config)s."""

from typing import Iterator
import copy
from typing import Generator, Union

from vizier._src.pyvizier.shared.parameter_config import ParameterConfig
from vizier._src.pyvizier.shared.parameter_config import SearchSpace
from vizier._src.pyvizier.shared.trial import ParameterDict
from vizier._src.pyvizier.shared.trial import ParameterValueTypes


class SequentialParameterBuilder(Iterator[ParameterConfig]):
  """Builds a ParameterDict by choosing one parameter value at a time.

  Example usage:
    from vizier import pyvizier as vz

    def decide_value(pc: vz.ParameterConfig) -> vz.ParameterValueTypes:
      ...

    search_space : vz.SearchSpace
    builder = SequentialParameterBuilder(search_space)

    for pc in builder:
      builder.choose_value(decide_value(pc))

    assert isinstance(builder.parameters, vz.ParameterDict)
  """

  def __init__(self,
               search_space: SearchSpace,
              
               *,
               traverse_order: str = 'dfs'):
    """Init.

    See the class pydoc for more details.

    Args:
      search_space: Search space to iterate over.
      traverse_order: 'dfs' or 'bfs'.
    """
    self._parameters = ParameterDict()
    self._traverse_order = traverse_order
    self._gen = self._coroutine(search_space)
    self._next = next(self._gen)
    self._stop_iteration = None

  def _coroutine(
      self, search_space: SearchSpace
  ) -> Generator[ParameterConfig, Union[ParameterValueTypes, None], None]:
    search_space = copy.deepcopy(search_space)
    while search_space.parameters:
      parameter_config = search_space.parameters[0]
      value = yield parameter_config
      if value is None:
        search_space.pop(parameter_config.name)
        continue
      subspace = search_space.get(parameter_config.name).get_subspace_deepcopy(
          value
      )
      search_space.pop(parameter_config.name)
      self._parameters[parameter_config.name] = value

      if self._traverse_order == 'bfs':
        # For BFS: append the subspace to the current search space.
        for child_parameter in subspace.parameters:
          search_space.add(child_parameter)
      else:
        # For DFS: append the current search space to the subspace.
        for parameter in search_space.parameters:
          subspace.add(parameter)
        search_space = subspace

  def __next__(self) -> ParameterConfig:
    if self._stop_iteration is not None:
      raise self._stop_iteration
    return self._next

  def choose_value(self, value: ParameterValueTypes) -> None:
    """Choose the value for the last ParameterConfig."""
    try:
      self._next = self._gen.send(value)
    except StopIteration as e:
      self._stop_iteration = e

  def skip(self) -> None:
    """Skip the value for the last ParameterConfig."""
    try:
      self._next = self._gen.send(None)
    except StopIteration as e:
      self._stop_iteration = e

  @property
  def parameters(self) -> ParameterDict:
    """Parameters chosen so far.

    WARNING: Do not mutate the dict until this Iterator is exhausted.
    """
    return self._parameters



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Utils for attrs."""

import re
from typing import Any, Callable, Collection, Optional

import attr

_Validator = Callable[[Any, attr.Attribute, Any], None]


def assert_not_empty(instance: Any, attribute: attr.Attribute,
                     value: Any) -> None:
  if not value:
    raise ValueError(f'{attribute.name} must not be empty in {type(instance)}.')


def assert_not_negative(instance: Any, attribute: attr.Attribute,
                        value: int) -> None:
  if value < 0:
    raise ValueError(
        f'{attribute.name} must not be negative in {type(instance)}.')


def assert_not_none(instance: Any, attribute: attr.Attribute,
                    value: Any) -> None:
  if value is None:
    raise ValueError(f'{attribute.name} must not be None in {type(instance)}.')


def assert_between(
    low: float, high: float
) -> Callable[[Any, attr.Attribute, str], None]:
  def validator(instance, attribute, value):
    del instance
    if value < low or value > high:
      raise ValueError(
          f'{attribute.name} ({value}) must be between {low} and {high}'
      )

  return validator


def assert_re_fullmatch(
    regex: str) -> Callable[[Any, attr.Attribute, str], None]:

  def validator(instance: Any, attribute: attr.Attribute, value: str) -> None:
    if not re.fullmatch(regex, value):
      raise ValueError(
          f'{attribute.name} must match the regex {regex} in {type(instance)}.')

  return validator


def shape_equals(instance_to_shape: Callable[[Any], Collection[Optional[int]]]):
  """Creates a shape validator for attrs.

  For example, _shape_equals(lambda s : [3, None]) validates that the shape has
  length 2 and its first element is 3.

  Code Example:
  @attrs.define
  class TestAttr:
    x = attrs.field(validator=attrs_utils.shape_equals(lambda v: (3, v.d)))
    d = attrs.field()

  _TestAttr(np.zeros([3, 2]), 2)  # OK
  _TestAttr(np.zeros([3, 5]), None) # OK
  _TestAttr(np.zeros([3, 2]), 4)  # Raises ValueError

  Args:
    instance_to_shape: Takes instance as input and returns the desired shape for
      the instance. `None` is treated as "any number".

  Returns:
    A validator that can be passed into attrs.ib or attrs.field.
  """

  def validator(instance, attribute, value) -> None:
    shape = instance_to_shape(instance)

    def _validator_boolean():
      if len(value.shape) != len(shape):
        return False
      for s1, s2 in zip(value.shape, shape):
        if (s2 is not None) and (s1 != s2):
          return False
      return True

    if not _validator_boolean():
      raise ValueError(f'{attribute.name} has shape {value.shape} '
                       f'which does not match the expected shape {shape}')

  return validator



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Tests for input_warping."""

import numpy as np
from vizier import pyvizier as vz
from vizier.pyvizier.converters import core as converters
from vizier.pyvizier.converters import input_warping

from absl.testing import absltest


class InputWarpingTest(absltest.TestCase):

  def test_kumaraswamy_transformation(self):
    rng = np.random.default_rng(seed=0)
    for a in np.linspace(0.01, 2.0, 50):
      for b in np.linspace(0.01, 2.0, 50):
        x = rng.uniform(size=(50, 20))
        f = input_warping.kumaraswamy_cdf(x, a, b)
        # Test that all values are in [0,1]
        self.assertEqual(np.sum((0 <= f) & (f <= 1)), x.shape[0] * x.shape[1])
        # Test the inverse CDF.
        np.testing.assert_allclose(
            input_warping.kumaraswamy_inv_cdf(f, a, b), x, atol=1e-07
        )

  def test_convert_trials(self):
    rng = np.random.default_rng(seed=0)
    search_space = vz.SearchSpace()
    for i in range(16):
      search_space.root.add_float_param(f"x{i}", 0.0, 1.0)
    problem = vz.ProblemStatement(search_space)
    converter = converters.TrialToArrayConverter.from_study_config(problem)
    input_warper = input_warping.KumaraswamyInputWarpingConverter(
        converter, a=0.1, b=0.8
    )

    trial1 = vz.Trial(parameters={f"x{i}": rng.uniform() for i in range(16)})
    trial2 = vz.Trial(parameters={f"x{i}": rng.uniform() for i in range(16)})
    trial3 = vz.Trial(parameters={f"x{i}": rng.uniform() for i in range(16)})
    trials = [trial1, trial2, trial3]

    features = input_warper.to_features(trials)
    self.assertEqual(np.sum((0 <= features) & (features <= 1)), 3 * 16)
    parameters = input_warper.to_parameters(features)
    for trial_idx, trial in enumerate(trials):
      for i in range(16):
        self.assertAlmostEqual(
            parameters[trial_idx][f"x{i}"].value,
            trial.parameters[f"x{i}"].value,
        )


if __name__ == "__main__":
  absltest.main()



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Tests for vizier.pythia.base.local_policy_supporters."""

import numpy as np
from vizier import pyvizier as vz
from vizier._src.pythia import local_policy_supporters
from absl.testing import absltest
from absl.testing import parameterized

InRamPolicySupporter = local_policy_supporters.InRamPolicySupporter
_GUID = '31'


def _runner_with_10trials():
  runner = InRamPolicySupporter(vz.ProblemStatement(), study_guid=_GUID)
  runner.AddTrials([vz.Trial() for _ in range(10)])
  return runner


class LocalPolicySupportersTest(parameterized.TestCase):

  def test_time_remaining(self):
    runner = _runner_with_10trials()
    runner.TimeRemaining()

  def test_add_and_get_trials(self):
    runner = _runner_with_10trials()
    trials = runner.GetTrials()
    self.assertLen(trials, 10)
    # The 10 trials are assigned ids 1 through 10 automatically.
    self.assertSequenceEqual([t.id for t in trials], range(1, 11))

    # Add 5 more Trials.
    runner = _runner_with_10trials()
    runner.AddTrials([vz.Trial(id=-1) for _ in range(5)])
    trials = runner.GetTrials()
    self.assertLen(trials, 15)
    self.assertSequenceEqual([t.id for t in trials], range(1, 16))

  def test_add_completed_trials(self):
    runner = _runner_with_10trials()
    trials = runner.GetTrials()

    trial = trials[0]
    trial.complete(vz.Measurement(), infeasibility_reason='test')
    runner.AddTrials([trial])
    # Second call should be a no-op.
    runner.AddTrials([trial])

    trials = runner.GetTrials()
    self.assertLen(trials, 10)
    # The 10 trials are assigned ids 1 through 10 automatically.
    self.assertSequenceEqual([t.id for t in trials], range(1, 11))
    self.assertEqual(trials[0].infeasibility_reason, 'test')

    trials = runner.GetTrials(status_matches=vz.TrialStatus.ACTIVE)
    self.assertLen(trials, 9)

  def test_prior_studies(self):
    runner = _runner_with_10trials()

    # Add a PriorStudy with study guid 1 and 10 completed Trials.
    prior_runner = _runner_with_10trials()
    completed_trials = []
    for t in prior_runner.GetTrials():
      t.complete(vz.Measurement())
      completed_trials.append(t)
    study = vz.ProblemAndTrials(
        problem=vz.ProblemStatement(), trials=completed_trials
    )
    study_guid = runner.SetPriorStudy(study, study_guid='1')
    self.assertEqual(study_guid, '1')

    # Add another PriorStudy with generated study guid and it has
    # 5 active and completed Trials
    prior_runner = _runner_with_10trials()
    trials = []
    for idx, t in enumerate(prior_runner.GetTrials()):
      if idx > 4:
        t.complete(vz.Measurement())
      trials.append(t)
    study = vz.ProblemAndTrials(problem=vz.ProblemStatement(), trials=trials)
    generated_study_guid = runner.SetPriorStudy(study)

    active_trials = runner.GetTrials(
        study_guid=study_guid, status_matches=vz.TrialStatus.ACTIVE
    )
    self.assertEmpty(active_trials)
    completed_trials = runner.GetTrials(
        study_guid=study_guid, status_matches=vz.TrialStatus.COMPLETED
    )
    self.assertLen(completed_trials, 10)

    active_trials = runner.GetTrials(
        study_guid=generated_study_guid, status_matches=vz.TrialStatus.ACTIVE
    )
    self.assertLen(active_trials, 5)
    completed_trials = runner.GetTrials(
        study_guid=generated_study_guid, status_matches=vz.TrialStatus.COMPLETED
    )
    self.assertLen(completed_trials, 5)

    with self.assertRaises(KeyError):
      runner.GetTrials(study_guid='2')

  def test_push_metadata(self):
    runner = _runner_with_10trials()
    trial1 = runner.GetTrials(min_trial_id=1, max_trial_id=1)[0]

    mu = vz.MetadataDelta()
    mu.assign('ns', 'key', 'value')
    mu.assign('ns', 'key', 'value', trial_id=1)
    # Metadata update is not immediate.
    self.assertEmpty(runner.GetStudyConfig(study_guid=_GUID).metadata.ns('ns'))
    self.assertEmpty(trial1.metadata)
    runner._UpdateMetadata(mu)

    self.assertEqual(
        runner.GetStudyConfig(study_guid=_GUID).metadata.ns('ns').get('key'),
        'value')
    trial0 = runner.GetTrials(min_trial_id=1, max_trial_id=1)[0]
    self.assertSequenceEqual(trial0.metadata.ns('ns'), {'key': 'value'})

  # TODO: Need a test for LocalPolicySupporter.SuggestTrials().


class LocalPolicySupportersGetBestTrialsTest(parameterized.TestCase):

  @parameterized.parameters(
      dict(goal=vz.ObjectiveMetricGoal.MAXIMIZE, count=2, best_values=[9, 8]),
      dict(goal=vz.ObjectiveMetricGoal.MINIMIZE, count=2, best_values=[0, 1]),
      dict(goal=vz.ObjectiveMetricGoal.MINIMIZE, count=None, best_values=[
          0,
      ]))
  def test_get_best_trials_single_objective(self, goal, count, best_values):
    runner = InRamPolicySupporter(
        vz.ProblemStatement(
            vz.SearchSpace(),
            metric_information=vz.MetricsConfig(
                [vz.MetricInformation(name='objective', goal=goal)])))

    trials = [
        vz.Trial().complete(vz.Measurement({'objective': i})) for i in range(10)
    ]
    np.random.shuffle(trials)
    runner.AddTrials(trials)

    objectives = np.array([
        t.final_measurement.metrics['objective'].value
        for t in runner.GetBestTrials(count=count)
    ])
    np.testing.assert_array_equal(objectives, np.array(best_values))

  @parameterized.parameters(
      dict(
          sin_goal=vz.ObjectiveMetricGoal.MAXIMIZE,
          cos_goal=vz.ObjectiveMetricGoal.MAXIMIZE,
          best_r=5),
      dict(
          sin_goal=vz.ObjectiveMetricGoal.MINIMIZE,
          cos_goal=vz.ObjectiveMetricGoal.MINIMIZE,
          best_r=1))
  def test_get_best_trials_multi_objective(self, sin_goal, cos_goal, best_r):
    runner = InRamPolicySupporter(
        vz.ProblemStatement(
            vz.SearchSpace(),
            metric_information=vz.MetricsConfig([
                vz.MetricInformation(name='sin', goal=sin_goal),
                vz.MetricInformation(name='cos', goal=cos_goal)
            ])))

    def build_measurement(r, theta):
      return vz.Measurement({
          'r': r,
          'theta': theta,
          'cos': r * np.cos(theta),
          'sin': r * np.sin(theta)
      })

    # Generate many trials for each radius.
    for r in range(1, 6):
      for theta in np.linspace(0, np.pi / 2, 5):
        runner.AddTrials([vz.Trial().complete(build_measurement(r, theta))])

    # Check the radius.
    rs = np.array([
        t.final_measurement.metrics['r'].value for t in runner.GetBestTrials()
    ])
    self.assertEqual(rs.size, 5)
    np.testing.assert_array_equal(rs, np.ones_like(rs) * best_r)

  @parameterized.parameters(
      dict(
          sin_goal=vz.ObjectiveMetricGoal.MAXIMIZE,
          cos_goal=vz.ObjectiveMetricGoal.MAXIMIZE,
      ),
      dict(
          sin_goal=vz.ObjectiveMetricGoal.MINIMIZE,
          cos_goal=vz.ObjectiveMetricGoal.MINIMIZE,
      ),
  )
  def test_get_best_trials_safe(self, sin_goal, cos_goal):
    runner = InRamPolicySupporter(
        vz.ProblemStatement(
            vz.SearchSpace(),
            metric_information=vz.MetricsConfig([
                vz.MetricInformation(name='sin', goal=sin_goal),
                vz.MetricInformation(
                    name='cos', goal=cos_goal, safety_threshold=0.51
                ),
            ]),
        )
    )

    def build_measurement(r, theta):
      return vz.Measurement({
          'r': r,
          'theta': theta,
          'cos': float(r * np.cos(theta)),
          'sin': float(r * np.sin(theta)),
      })

    # Generate many trials for each radius.
    for r in range(1, 6):
      for theta in np.linspace(0, np.pi / 2, 5):
        runner.AddTrials([vz.Trial().complete(build_measurement(r, theta))])

    # Check the radius.
    best_trials = runner.GetBestTrials()
    self.assertLen(best_trials, 1)
    cosine = best_trials[0].final_measurement.metrics['cos'].value
    if cos_goal == vz.ObjectiveMetricGoal.MAXIMIZE:
      self.assertGreaterEqual(cosine, 0.51)
    else:
      self.assertLessEqual(cosine, 0.51)


if __name__ == '__main__':
  absltest.main()



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Tests for optimizers."""

import datetime

from absl.testing import parameterized
import jax
from tensorflow_probability.substrates import jax as tfp
from vizier._src.jax.optimizers import jaxopt_wrappers as jw
from vizier._src.jax.optimizers.testing import sinusoidal

from absl.testing import absltest

tfb = tfp.bijectors


class JaxoptWrappersTest(
    parameterized.TestCase, sinusoidal.ConvergenceTestMixin
):

  @parameterized.product(
      (
          dict(bounds=None, nest_constraint=True),
          dict(bounds=((-4.0, None)), nest_constraint=True),
          dict(bounds=((None, 5.0)), nest_constraint=False),
          dict(bounds=((-3.0, 3.0)), nest_constraint=True),
      ),
      cls=(jw.JaxoptScipyLbfgsB, jw.JaxoptLbfgsB),
  )
  def test_sinusoidal(self, cls, bounds, nest_constraint):
    constraints = sinusoidal.bounds_to_constraints(bounds, nest_constraint)
    self.assert_converges(
        cls(jw.LbfgsBOptions()),
        constraints=constraints,
        threshold=1.0,
        random_restarts=20,
    )

  def test_max_duration(self):
    optimizer = jw.JaxoptScipyLbfgsB(
        max_duration=datetime.timedelta(seconds=0),
        speed_test=True,
    )
    random_restarts = 3
    rngs = jax.random.split(jax.random.PRNGKey(1), random_restarts + 1)
    _, metrics = optimizer(
        init_params=jax.vmap(sinusoidal.setup)(rngs[1:]),
        loss_fn=jax.jit(sinusoidal.loss_fn),
        rng=rngs[0],
    )
    # Check that there's only one training instead of 3.
    self.assertLen(metrics['train_time'], 1)


if __name__ == '__main__':
  absltest.main()



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Separate Pythia service for handling algorithmic logic."""
from typing import Optional

from absl import logging
import attr
import grpc
from vizier import pythia
from vizier._src.service import policy_factory as service_policy_factory
from vizier._src.service import pythia_service_pb2
from vizier._src.service import pythia_service_pb2_grpc
from vizier._src.service import service_policy_supporter
from vizier._src.service import stubs_util
from vizier._src.service import types
from vizier.service import pyvizier as vz

from google.protobuf import empty_pb2


@attr.define
class PythiaServicer(pythia_service_pb2_grpc.PythiaServiceServicer):
  """Implements the GRPC functions outlined in pythia_service.proto."""

  # Can be either an actual VizierService object or a stub. An actual
  # VizierService should be passed only for local testing/local development.
  _vizier_service: Optional[types.VizierService] = attr.field(
      init=True, default=None
  )
  # Factory for creating policies. Defaulted to OSS Vizier-specific policies,
  # but allows use of external package (e.g. PyGlove) poliices.
  _policy_factory: pythia.PolicyFactory = attr.field(
      init=True, factory=service_policy_factory.DefaultPolicyFactory
  )

  def __attrs_post_init__(self):
    try:
      # If Jax is installed, always use float64 for all policies.
      import jax  # pylint:disable=g-import-not-at-top

      jax.config.update('jax_enable_x64', True)
    except ImportError:
      pass

  def connect_to_vizier(self, endpoint: str) -> None:
    """Only needs to be called if VizierService wasn't passed in init."""
    if self._vizier_service:
      raise ValueError('Vizier Service was already set:', self._vizier_service)
    self._vizier_service = stubs_util.create_vizier_server_stub(endpoint)

  def Suggest(
      self,
      request: pythia_service_pb2.SuggestRequest,
      context: Optional[grpc.ServicerContext] = None,
  ) -> pythia_service_pb2.SuggestDecision:
    """Performs Suggest RPC call."""
    # Setup Policy Supporter.
    study_config = vz.SuggestConverter.from_request_proto(request).study_config
    study_name = request.study_descriptor.guid
    policy_supporter = service_policy_supporter.ServicePolicySupporter(
        study_name, self._vizier_service
    )
    pythia_policy = self._policy_factory(
        study_config, request.algorithm, policy_supporter, study_name
    )

    # Perform algorithmic computation.
    suggest_request = vz.SuggestConverter.from_request_proto(request)
    try:
      suggest_decision = pythia_policy.suggest(suggest_request)
    # Leaving a broad catch for now since Pythia can raise any exception.
    # TODO: Be more specific about exception raised,
    # e.g. AttributeError, ModuleNotFoundError, SyntaxError
    except Exception as e:  # pylint: disable=broad-except
      logging.error(
          'Failed to request trials from Pythia for request: %s', request
      )
      raise RuntimeError('Pythia has encountered an error: ' + str(e)) from e

    return vz.SuggestConverter.to_decision_proto(suggest_decision)

  def EarlyStop(
      self,
      request: pythia_service_pb2.EarlyStopRequest,
      context: Optional[grpc.ServicerContext] = None,
  ) -> pythia_service_pb2.EarlyStopDecisions:
    """Performs EarlyStop RPC call."""
    # Setup Policy Supporter.
    study_config = vz.EarlyStopConverter.from_request_proto(
        request
    ).study_config
    study_name = request.study_descriptor.guid
    policy_supporter = service_policy_supporter.ServicePolicySupporter(
        study_name, self._vizier_service
    )
    pythia_policy = self._policy_factory(
        study_config, request.algorithm, policy_supporter, study_name
    )

    # Perform algorithmic computation.
    early_stop_request = vz.EarlyStopConverter.from_request_proto(request)
    early_stopping_decisions = pythia_policy.early_stop(early_stop_request)

    return vz.EarlyStopConverter.to_decisions_proto(early_stopping_decisions)

  def Ping(
      self,
      request: empty_pb2.Empty,
      context: Optional[grpc.ServicerContext] = None,
  ) -> empty_pb2.Empty:
    return empty_pb2.Empty()



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""SimpleKD experimenter.

The experimenter supports three flavors (corner, center, mixed), each of which
has a different optia location. The categorical parameter takes the same three
values, and the objective function depends on the value of the categorical
parameter which make it harder to optimize.
"""

, Sequence, Union
import attrs
import numpy as np
from vizier import pyvizier as vz
from vizier._src.benchmarks.experimenters import experimenter


def _float_term(x_list: list[float]) -> float:
  """Computes the float term on a list of values.

  Args:
    x_list: Elements in the list correspond to a different dimension/Parameter.

  Returns:
    The float term accounting for all the float Parameters.
  """
  float_term = 0
  for x in x_list:
    float_term += min(
        1.6,
        (x > -0.8) * -((x - 1) ** 2)
        + (x <= -0.8) * (-(1.8**2) + 150 * (x + 0.8) ** 2),
    )
  return float_term


SimpleKDCategory = str


def _categorical_term(x: str, best_category: SimpleKDCategory) -> float:
  """Computes the categorical term."""
  if x != best_category:
    return 0
  elif x == 'corner':
    return 1
  elif x == 'center':
    return 1
  elif x == 'mixed':
    return 1.5
  raise NotImplementedError(f'Unknown categorical parameter: {x}')


_feasible_discrete_values = (1, 2, 5, 6, 8)


def _discrete_term(x_list: list[int]) -> float:
  """Computes the discrete term on a list of values."""
  discrete_term = 0
  for x in x_list:
    discrete_term += [1.2, 0.0, 0.6, 0.8, 1.0][
        _feasible_discrete_values.index(x)
    ]
  return discrete_term


def _int_term(x_list: list[int]) -> float:
  """Computes the int term on a list of values."""
  int_term = 0
  for x in x_list:
    int_term += np.power(x - 2.2, 2) / 2.0
  return int_term


@attrs.define
class SimpleKD(experimenter.Experimenter):
  """Simple problem with arbitrary number of parameter of each type.

  SimpleKD is a carefully constructed function to trap an underexploring
  hillclimb algorithm to local optima. It is parameterized by "best_category",
  which optimization algorithm must get right in order to reach the optimum.

  Note that CATEGORICAL type only has one parameter as order doesn't exist.

  "best_category" value refers to where the optimum occurs.
  """

  best_category: SimpleKDCategory = attrs.field()  # type: ignore
  num_float_param: int = 1
  num_discrete_param: int = 1
  num_int_param: int = 1

  def evaluate(self, suggestions: Sequence[vz.Trial]) -> None:
    for suggestion in suggestions:
      values = self._get_param_values(suggestion)
      suggestion.complete(vz.Measurement({'value': self._compute(values)}))

  def problem_statement(self) -> vz.ProblemStatement:
    problem = vz.ProblemStatement()
    for f in range(self.num_float_param):
      problem.search_space.root.add_float_param(f'float_{f}', -1, 1)
    for d in range(self.num_discrete_param):
      problem.search_space.root.add_discrete_param(
          f'discrete_{d}', _feasible_discrete_values
      )
    for i in range(self.num_int_param):
      problem.search_space.root.add_int_param(f'int_{i}', 1, 3)
    problem.search_space.root.add_categorical_param(
        'categorical', ('corner', 'center', 'mixed')
    )
    problem.metric_information.append(
        vz.MetricInformation(name='value', goal=vz.ObjectiveMetricGoal.MAXIMIZE)
    )
    return problem

  def _get_param_values(
      self, trial: vz.Trial
  ) -> dict[str, list[Union[float, int, str]]]:
    """Extract the trial parameter values by type."""

    float_list = []
    int_list = []
    discrete_list = []
    for index in range(self.num_float_param):
      float_list.append(trial.parameters[f'float_{index}'].value)
    for index in range(self.num_discrete_param):
      discrete_list.append(trial.parameters[f'discrete_{index}'].value)
    for index in range(self.num_int_param):
      int_list.append(trial.parameters[f'int_{index}'].value)

    return {
        'float': float_list,
        'int': int_list,
        'discrete': discrete_list,
        'categorical': [trial.parameters['categorical'].value],
    }

  def _compute(self, params: dict[str, list[Union[float, int, str]]]) -> float:
    """Computes the SimpleKD objective value."""
    if params['categorical'][0] == 'corner':
      return (
          _categorical_term(params['categorical'][0], self.best_category)
          + 0.8 * _float_term(params['float'])
          + _discrete_term(params['discrete'])
          + _int_term(params['int'])
      )
    elif params['categorical'][0] == 'center':
      return (
          _categorical_term(params['categorical'][0], self.best_category)
          - _float_term(params['float'])
          - _discrete_term(params['discrete'])
          - _int_term(params['int'])
      )
    elif params['categorical'][0] == 'mixed':
      return (
          _categorical_term(params['categorical'][0], self.best_category)
          + 0.8 * _float_term(params['float'])
          + _discrete_term(params['discrete'])
          - _int_term(params['int'])
      )
    else:
      raise NotImplementedError(
          f'Unknown categorical parameter: {params["categorical"]}'
      )

  @property
  def optimal_objective(self) -> float:
    """Computes the optimal objective value of a SimpleKD problem."""
    values = {}
    if self.best_category == 'corner':
      values['float'] = [-1.0 for _ in range(self.num_float_param)]
      values['int'] = [1 for _ in range(self.num_float_param)]
      values['discrete'] = [1 for _ in range(self.num_float_param)]
      values['categorical'] = ['corner']

    if self.best_category == 'center':
      values['float'] = [-0.8 for _ in range(self.num_float_param)]
      values['int'] = [2 for _ in range(self.num_float_param)]
      values['discrete'] = [2 for _ in range(self.num_float_param)]
      values['categorical'] = ['center']

    if self.best_category == 'mixed':
      values['float'] = [-1.0 for _ in range(self.num_float_param)]
      values['int'] = [2 for _ in range(self.num_float_param)]
      values['discrete'] = [1 for _ in range(self.num_float_param)]
      values['categorical'] = ['mixed']

    return self._compute(values)



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Tests for clients."""
import functools

from absl import logging
from vizier._src.service import clients
from vizier._src.service import constants
from vizier._src.service import vizier_server
from vizier.client import client_abc_testing
from vizier.service import pyvizier as vz

from absl.testing import absltest


# Affects local Vizier servicer tests only.
clients.environment_variables.servicer_use_sql_ram()


class VizierClientTest(client_abc_testing.TestCase):
  _owner: str

  def create_study(
      self, problem: vz.ProblemStatement, study_id: str
  ) -> clients.Study:
    config = vz.StudyConfig.from_problem(problem)
    config.algorithm = vz.Algorithm.RANDOM_SEARCH
    study = clients.Study.from_study_config(
        config, owner='owner', study_id=study_id
    )
    return study

  def test_e2e_tuning(self):
    self.assertPassesE2ETuning()

  def test_e2e_tuning_gp(self):
    def create_study(
        problem: vz.ProblemStatement, study_id: str
    ) -> clients.Study:
      config = vz.StudyConfig.from_problem(problem)
      config.algorithm = vz.Algorithm.GAUSSIAN_PROCESS_BANDIT
      study = clients.Study.from_study_config(
          config, owner='owner', study_id=study_id
      )
      return study

    study_factory = functools.partial(create_study, study_id=self.id())
    self.assertPassesE2ETuning(study_factory=study_factory, num_iterations=2)


class VizierClientTestOnServicer(VizierClientTest):

  @classmethod
  def setUpClass(cls):
    logging.info('Test setup started.')
    super().setUpClass()
    clients.environment_variables.server_endpoint = constants.NO_ENDPOINT
    logging.info('Test setup finished.')


class VizierClientTestOnDefaultServer(VizierClientTest):
  _server: vizier_server.DefaultVizierServer

  @classmethod
  def setUpClass(cls):
    logging.info('Test setup started.')
    super().setUpClass()
    cls._server = vizier_server.DefaultVizierServer(
        database_url=constants.SQL_MEMORY_URL
    )
    clients.environment_variables.server_endpoint = cls._server.endpoint
    logging.info('Test setup finished.')

  @classmethod
  def tearDownClass(cls):
    cls._server._server.stop(None)
    super().tearDownClass()


class VizierClientTestOnDistributedPythiaServer(VizierClientTest):
  _server: vizier_server.DistributedPythiaVizierServer

  @classmethod
  def setUpClass(cls):
    logging.info('Test setup started.')
    super().setUpClass()
    cls._server = vizier_server.DistributedPythiaVizierServer(
        database_url=constants.SQL_MEMORY_URL
    )
    clients.environment_variables.server_endpoint = cls._server.endpoint
    logging.info('Test setup finished.')

  @classmethod
  def tearDownClass(cls):
    cls._server._server.stop(None)
    cls._server._pythia_server.stop(None)
    super().tearDownClass()


if __name__ == '__main__':
  absltest.main()



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Tests for acquisitions."""

import functools

import jax
from jax import config
import numpy as np
from tensorflow_probability.substrates import jax as tfp
from vizier import pyvizier as vz
from vizier._src.algorithms.designers.gp import acquisitions
from vizier._src.jax import gp_bandit_utils
from vizier._src.jax import predictive_fns
from vizier._src.jax import stochastic_process_model as sp
from vizier._src.jax import types
from vizier._src.jax.models import tuned_gp_models
from vizier.jax import optimizers

from absl.testing import absltest


tfd = tfp.distributions


class GPBanditAcquisitionBuilderTest(absltest.TestCase):

  def test_categorical_kernel(self, best_n=2):
    # Random key
    key = jax.random.PRNGKey(0)
    # Simulate data
    n_samples = 10
    n_continuous = 3
    n_categorical = 5
    features = types.ModelInput(
        continuous=types.PaddedArray.as_padded(
            jax.random.normal(key, shape=(n_samples, n_continuous)),
        ),
        categorical=types.PaddedArray.as_padded(
            jax.random.normal(key, shape=(n_samples, n_categorical)),
        ),
    )
    labels = types.PaddedArray.as_padded(
        jax.random.normal(key, shape=(n_samples, 1)),
    )
    xs = features
    # Model
    coroutine = tuned_gp_models.VizierGaussianProcess(
        _dim=types.ContinuousAndCategorical[int](n_continuous, n_categorical)
    )
    model = sp.StochasticProcessModel(coroutine)
    loss_fn = functools.partial(
        jax.jit(
            gp_bandit_utils.stochastic_process_model_loss_fn,
            static_argnames=('model', 'normalize'),
        ),
        model=model,
        data=types.ModelData(features=features, labels=labels),
    )
    setup = lambda rng: model.init(rng, features)['params']
    constraints = sp.get_constraints(model)

    # ARD
    ard_optimizer = optimizers.JaxoptScipyLbfgsB()
    key, init_key = jax.random.split(key, 2)
    best_model_params, _ = ard_optimizer(
        init_params=jax.vmap(setup)(jax.random.split(init_key, 4)),
        loss_fn=loss_fn,
        rng=key,
        constraints=constraints,
        best_n=best_n,
    )

    def precompute_cholesky(params):
      _, pp_state = model.apply(
          {'params': params},
          features,
          labels,
          method=model.precompute_predictive,
          mutable='predictive',
      )
      return pp_state

    pp_state = jax.vmap(precompute_cholesky)(best_model_params)

    # Create the state.
    model_state = {'params': best_model_params, **pp_state}
    # Define the problem.
    space = vz.SearchSpace()
    root = space.root
    for j in range(n_continuous):
      root.add_float_param(f'f{j}', -1.0, 2.0)
    for j in range(n_categorical):
      root.add_categorical_param(f'c{j}', ['a', 'b', 'c'])
    problem = vz.ProblemStatement(space)
    problem.metric_information.append(
        vz.MetricInformation(
            name='metric', goal=vz.ObjectiveMetricGoal.MAXIMIZE
        )
    )
    data = types.ModelData(features=features, labels=labels)
    state = types.GPState(model_state=model_state, data=data)
    pred_dict = predictive_fns.predict_on_array(xs, model=model, state=state)

    acquisition_val = predictive_fns.acquisition_on_array(
        xs, model, acquisition_fn=acquisitions.UCB(), state=state
    )
    self.assertFalse(np.any(np.isnan(pred_dict['mean'])))
    self.assertFalse(np.any(np.isnan(pred_dict['stddev'])))
    self.assertFalse(np.any(np.isnan(acquisition_val)))


if __name__ == '__main__':
  config.update('jax_enable_x64', True)
  absltest.main()
